def get_exception(instance):
    """
    Method to find current class Exception
    :param instance: instance of a classe having an exception defined in
    estrade.exception
    :return: <(Class)Exception>
    """
    module = __import__('estrade.exceptions', fromlist=['estrade.exceptions'])
    class_name = f'{instance.__class__.__name__}Exception'
    return getattr(module, class_name)



""" This file define the Candle class.
"""
import logging
from typing import Any, Dict, List, Optional

from estrade.exceptions import CandleException
from estrade.tick import Tick

logger = logging.getLogger(__name__)


class Candle:
    """
    Class used to represent a Candle (from a clandlesticks graph).
    This class is used in list of candles of `estrade.candle_set.CandleSet` instances.

    A candle holds a list of ticks happening between the time/nb defined in CandleSet.

    Arguments:
        timeframe: candle timeframe
        epic_ref: reference of epic
        meta: free of use dictionnary

    !!! note
        The `open_at` can be different from the open_tick datetime, for example in a
        `5minutes` timeframe, if the first tick has a time of 12:06:03, then the
        `open_at` should be 12:05:00.

    !!! note
        This class does not have a reference to a CandleSet because it can be created
        manually from a Provider (and not be attached to a candleset directly).

    """

    def __init__(
        self,
        timeframe: str,
        epic_ref: str,
        open_tick: Tick,
        meta: Dict[str, Any] = None,
    ) -> None:
        logger.debug('create new candle')

        # validate and set timeframe
        from estrade.candle_set import CandleSet

        CandleSet.validate_timeframe(timeframe)
        self.timeframe = timeframe

        # TODO: instanciate with open_tick.epic.ref
        if not isinstance(epic_ref, str):
            raise CandleException('Invalid Candle Epic ref')
        self.epic_ref = epic_ref

        if not isinstance(open_tick, Tick):
            raise CandleException('Invalid candle open tick')
        self.ticks: List[Tick] = [open_tick]
        self.low_tick = open_tick
        self.high_tick = open_tick

        self.closed = False

        # as the open tick is not necessarily on the timeframe lowest, set it with
        # the CandleSet static method
        nb, ut = CandleSet.split_timeframe(self.timeframe)
        self.open_at = CandleSet.round_candle_open_datetime(
            tick_datetime=open_tick.datetime, nb=nb, ut=ut,
        )

        self.indicators: Dict[str, Any] = {}
        self.meta = meta or {}

        logger.debug('new candle created: %s' % self.open_at)

    ##################################################
    # TICKS
    ##################################################
    @property
    def open(self) -> float:
        """
        Returns:
            The first tick of candle
        """
        return self.ticks[0].value

    @property
    def last(self) -> float:
        """
        Returns:
            The last tick of candle
        """
        return self.ticks[-1].value

    @property
    def close(self) -> Optional[float]:
        """
        Returns:
            When candle is closed, this method send the last tick of candle
        """
        if self.closed:
            return self.last
        return None

    @property
    def high(self) -> float:
        """
        Returns:
            The highest tick in candle
        """
        return self.high_tick.value

    @property
    def low(self) -> float:
        """
        Returns:
            The tick with the lowest value of candle

        """
        return self.low_tick.value

    @property
    def open_tick(self) -> Tick:
        """
        Returns:
            The fist tick of candle
        """
        return self.ticks[0]

    @property
    def last_tick(self) -> Tick:
        """
        Returns:
            The last tick in candle

        """
        return self.ticks[-1]

    ##################################################
    # CANDLE PARTS
    ##################################################
    @property
    def color(self) -> str:
        """
        Returns:
            This method return the "color" of the candle.

         - if open < close: returns `green`
         - elif open > close: returns `red`
         - else (open == close): returns `black` (doji)

        """
        if self.last > self.open:
            return 'green'
        elif self.last < self.open:
            return 'red'
        return 'black'

    @property
    def height(self) -> float:
        """
        Returns:
            Height of candle (high - low)

        ```
                |  ┐
                |  |
                ┴  |
               | | |
               | | |<-- height
               | | |
                ┬  |
                |  |
                |  ┘
        ```
        """
        return self.high - self.low

    @property
    def body(self) -> float:
        """
        Returns:
            the absolute difference between open and close of the candle

        ```
                |
                |
                ┴  ┐
               | | |
               | | |<-- body
               | | |
                ┬  ┘
                |
                |
        ```
        """
        return abs(self.open - self.last)

    @property
    def head(self) -> float:
        """
        Returns:
            distance between the highest value and the highest from (open/last)
        ```
                |  ┐
                |  | <-- head
                ┴  ┘
               | |
               | |
               | |
                ┬
                |
                |
        ```
        """
        return self.high - (max(self.open, self.last))

    @property
    def tail(self) -> float:
        """
        Returns:
            distance between the lowest value and the lowest from (open/last)

        ```
                |
                |
                ┴
               | |
               | |
               | |
                ┬  ┐
                |  | <-- tail
                |  ┘
        ```
        """
        return min(self.open, self.last) - self.low

    ##################################################
    # EVENTS
    ##################################################
    def on_new_tick(self, tick: Tick) -> None:
        """
        This method append a new tick to the candle.ticks and update candle low/high.

        Arguments:
            tick: tick to add to candle
        """
        if not isinstance(tick, Tick):
            raise CandleException('Can only add Tick objects to candle')

        self.ticks.append(tick)

        if not self.low_tick or tick.value < self.low_tick.value:
            self.low_tick = tick
        if not self.high_tick or tick.value > self.high_tick.value:
            self.high_tick = tick

    def close_candle(self) -> None:
        """
        This method set the candle as closed.

        A candle is closed when :

         - the max number of tick in candle is reached (max number defined in "parent"
         CandleSet)
         - the candle time range is reached (time ranched defined in "parent" CandleSet)
        """
        self.closed = True

    def __str__(self) -> str:
        return (
            f'Candle : open {self.open}@{self.open_at}, '
            f'{"close" if self.closed else "last"} {self.ticks[-1].value}@'
            f'{self.ticks[-1].datetime}'
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            'timeframe': self.timeframe,
            'epic': self.epic_ref,
            'nb_ticks': len(self.ticks),
            'open': self.open_tick.to_json(),
            'high': self.high_tick.to_json() if self.high_tick else None,
            'low': self.low_tick.to_json() if self.low_tick else None,
            'closed': self.closed,
            'close': self.ticks[-1].to_json() if self.closed else None,
            'indicators': self.indicators,
            'meta': self.meta,
        }



from estrade.mixins.candle_set_indicator_mixin import CandleSetIndicatorMixin
import logging

logger = logging.getLogger(__name__)


class MovingAverageMixin(CandleSetIndicatorMixin):
    def __init__(self, periods: int, name: str) -> None:
        self.periods = periods
        self.sum_finished = 0
        self._sma = None
        CandleSetIndicatorMixin.__init__(self, name=name)

    def _post_on_new_tick(self, tick):
        pass

    def on_new_tick(self, tick):
        logger.debug('Update Moving average %s with %d' % (self.name, tick.value))
        if len(self.candle_set.candles) >= self.periods:
            self._sma = (tick.value + self.sum_finished) / self.periods
            logger.debug('indicator %s: set mm value to %d' % (self.name, self._sma))
            self.candle_set.candles[-1].indicators[self.name] = self._sma

        self._post_on_new_tick(tick)

    def on_new_candle(self, new_candle):
        pass

    def _post_candle_close(self, closed_candle):
        pass

    def on_candle_close(self, closed_candle):
        if self.candle_set.nb_closed_candles >= (self.periods - 1):
            self.sum_finished = sum(
                [c.close for c in self.candle_set.closed_candles(nb=(self.periods - 1))]
            )

            logger.debug('Update sum of finished candles to %d' % (self.sum_finished,))
        self._post_candle_close(closed_candle)


class SimpleMovingAverage(MovingAverageMixin):
    """
    Simple moving average indicator to add on a `estrade.CandleSet` object.

    Arguments:
        periods: nb of periods
        name: name of indicator

    !!! warning
        It is **strongly** advised to give a name to your indicator so you can fetch
        their values in your strategy (see `estrade.Strategy.get_indicator`)

    !!! example
        ```python
        from estrade import Epic, CandleSet, Market, SimpleMovingAverage, Strategy

        class MyStrategy(Strategy):
            def on_new_candle_opening_strategy(candle_set):
                # fetch indicator value
                mm200 = candle_set.indicator(indicator_name='mm200')

                # check if indicator value is defined (for the first 199 ticks,
                # the Moving Average 200 will not be set)
                if mm200.value:
                    # do something....

            def on_new_tick_opening_strategy(tick):
                # to fetch indicator, the from a `on_new_tick...` method in strategy,
                # the easiest way is to use the `Strategy.get_indicator()` helper
                mm200 = self.get_indicator(timeframe='5minutes', name='mm200')

                # check if indicator value is defined (for the first 199 ticks,
                # the Moving Average 200 will not be set)
                if mm200.value:
                    # do something....


        if __name__ == '__main__':
            market = Market()
            epic = Epic(ref='MY_EPIC', market=market)
            mm200 = SimpleMovingAverage(periods=200, name='mm200')
            CandleSet(epic=epic, timeframe='5minutes', indicators=[mm200])
            # ...
            market.run()
        ```
    """

    @property
    def value(self):
        return round(self._sma, 2) if self._sma else None


class ExponentialMovingAverage(MovingAverageMixin):
    """
    Exponential moving average indicator to add on a `estrade.CandleSet` object.

    Arguments:
        periods: nb of periods
        name: name of indicator

    !!! warning
        It is **strongly** advised to give a name to your indicator so you can fetch
        their values in your strategy (see `estrade.Strategy.get_indicator`)

    !!! example
        ```python
        from estrade import Epic, CandleSet, ExponentialMovingAverage, Strategy

        class MyStrategy(Strategy):
            def on_new_candle_opening_strategy(candle_set):
                # fetch indicator value
                ema200 = candle_set.indicator(indicator_name='ema200')

                # check if indicator value is defined (for the first 199 ticks,
                # the Moving Average 200 will not be set)
                if ema200.value:
                    # do something....

            def on_new_tick_opening_strategy(tick):
                # to fetch indicator, the from a `on_new_tick...` method in strategy,
                # the easiest way is to use the `Strategy.get_indicator()` helper
                ema200 = self.get_indicator(timeframe='5minutes', name='ema200')

                # check if indicator value is defined (for the first 199 ticks,
                # the Moving Average 200 will not be set)
                if ema200.value:
                    # do something....


        if __name__ == '__main__':
            market = Market()
            epic = Epic(ref='MY_EPIC', market=market)
            ema200 = ExponentialMovingAverage(periods=200, name='ema200')
            CandleSet(epic=epic, timeframe='5minutes', indicators=[ema200])
            # ...
            market.run()
        ```

    """

    def __init__(self, periods: int, name: str) -> None:
        self._ema = None
        self.shooting_constant = 2 / (periods + 1)
        self.last_candle_sma = None
        MovingAverageMixin.__init__(self, periods=periods, name=name)

    @property
    def value(self):
        return round(self._ema, 2) if self._ema else None

    @property
    def sma(self):
        return round(self._sma, 2) if self._sma else None

    def _post_candle_close(self, _):
        if self._sma:
            self.last_candle_sma = self._sma

    def _post_on_new_tick(self, tick):
        if self.last_candle_sma:
            logger.debug(
                f'Calulate EMA: {self.shooting_constant} * '
                f'({tick.value} - {self.last_candle_sma}) + {self.last_candle_sma}'
            )
            self._ema = (
                self.shooting_constant * (tick.value - self.last_candle_sma)
                + self.last_candle_sma
            )



import logging

from typing import List, Optional, TYPE_CHECKING

from estrade.mixins.ref_mixin import RefMixin
from estrade.epic import Epic
from estrade.exceptions import MarketException
from estrade.observer import Observable
from estrade.strategy import Strategy
from estrade.tick import Tick
from estrade.trade_manager import TradeManager

if TYPE_CHECKING:
    from estrade.provider import Provider
    from estrade.mixins.reporting_mixin import ReportingMixin


logger = logging.getLogger(__name__)


class Market(RefMixin, Observable):
    """
    A market is the component wrapping provider and epics.

    Arguments:
        ref: name of the market

    !!! example
        ```python
        from estrade import Market

        market = Market(ref='MY_MARKET')
        ```
    """

    def __init__(self, ref: str = None) -> None:
        logger.info('Create new Market')

        # init ref via parent
        RefMixin.__init__(self, ref)

        # set as observable so market can fire events.
        Observable.__init__(self)

        # init epics as an empty list (will be filled by strategies epics)
        self.epics: List[Epic] = []
        self.strategies: List[Strategy] = []

        # Init new trade manager
        self.trade_manager: TradeManager = TradeManager(market=self)

        self.provider: Optional[Provider] = None

        self.reporting: Optional[ReportingMixin] = None

    ##################################################
    # EPICS
    ##################################################
    def get_epic(self, epic_ref: str) -> Epic:
        """
        Get an epic by its reference.

        Arguments:
            epic_ref: reference of epic to find

        Raises:
            MarketException: if no epic found with this reference

        Returns:
            epic instance
        """
        for e in self.epics:
            if e.ref == epic_ref:
                return e
        raise MarketException(f'Epic with ref {epic_ref} not found')

    @property
    def epics_refs(self) -> List[str]:
        """
        Get the list of epic references

        Returns:
            list of ref of this instance epics
        """
        return [epic.ref for epic in self.epics]

    ##################################################
    # STRATEGY
    ##################################################
    @property
    def one_strategy_still_active(self) -> bool:
        for strategy in self.strategies:
            if strategy.stopped is False:
                return True
        return False

    ##################################################
    # RUNTIME
    ##################################################
    def _dispatch_tick(self, tick: Tick) -> None:
        """
        Dispatch tick to :
         - epics (update of its candle sets & indicators)
         - strategies (open and close trades)

        """
        # dispatch tick to its epic to update candleset and indicators
        tick.epic.on_new_tick(tick=tick)

        # send tick to strategies attached to its epic
        for strategy in tick.epic.strategies:
            if strategy.stopped is False:
                logger.debug('dispatch tick to strategy %s' % strategy.ref)
                strategy.on_new_tick(tick=tick)

    def on_new_tick(self, tick: Tick) -> None:
        """
        This method is called every time a new tick is sent by provider.

        Its purpose is to:
         1. Update open trades with the new tick value
         2. Dispatch tick to epic (and update candle sets & indicators)
         3. Dispatch tick to strategies (to open and close trades)

        Arguments:
            tick: new tick to be dispatched to candle sets and strategies

        Raises:
            estrade.exceptions.TickException: if tick is not valid

        """
        logger.debug('new tick received: %s' % tick)

        Tick.validate(tick)

        # update open trades with the new tick
        self.trade_manager.on_new_tick(tick=tick)

        # dispatch tick to epics and strategies
        self._dispatch_tick(tick)

    def on_new_candle(self, candle):
        """
        This method is called every time a new candle is sent by self.provider.
        :param candle: :class:`estrade.candle.Candle`
        :return
        """
        candle_epic = candle.open_tick.epic
        epic_timeframes = [cs.timeframe for cs in candle_epic.candle_sets]
        if candle.timeframe not in epic_timeframes:
            raise MarketException(
                f'Impossible to add candle, Epic {candle_epic.ref} '
                f'has no CandleSet with timeframe {candle.timeframe}'
            )
        # dispatch open tick to epic candle_sets
        self.on_new_tick(tick=candle.open_tick)

        # dispatch high and low tick to ongoing trades
        # apply low first to open BUY and high first to open SELL
        self.trade_manager.on_new_ticks_high_low(
            epic=candle_epic, low=candle.low_tick, high=candle.high_tick,
        )
        # dispatch high/low to candle sets
        if candle.high_tick.datetime < candle.low_tick.datetime:
            self._dispatch_tick(candle.high_tick)
            self._dispatch_tick(candle.low_tick)
        else:
            self._dispatch_tick(candle.low_tick)
            self._dispatch_tick(candle.high_tick)

        # dispatch candle close
        self.on_new_tick(candle.ticks[-1])

    def run(self, mode: str = 'tick') -> None:
        """
        This method run the data Provider.
            - check provided is logged
            - check that no open trade exists in provider
            - generate ticks from provider (for every tick, the provider call the above
            method on_new_tick)
        """
        self.fire('market_before_run')

        if not self.provider:
            raise MarketException('Impossible to run Market without a provider')

        # add strategies to trade manager
        self.trade_manager.add_strategies()

        logger.info('Check if provider is logged')
        if self.provider.requires_login:
            if not self.provider.login():
                raise MarketException(
                    'Cannot start because provider is not logged, '
                    'please call the provider login function before running market.'
                )

        logger.info('Check opened trades')
        if self.provider.get_open_trades():
            raise MarketException('Cannot start when trades are already opened')

        logger.info('Generate ticks from provider')
        if mode == 'tick':
            for tick in self.provider.generate_ticks():
                self.on_new_tick(tick)
                if self.one_strategy_still_active is False:
                    break

        elif mode == 'candle':
            for candle in self.provider.generate_candles():
                self.on_new_candle(candle)
                if self.one_strategy_still_active is False:
                    break

        # trigger actions (reporting etc.) on market end
        self.fire('market_run_end')



import csv
import os

REPORT_FOLDER = 'reports'


class CSVWriter:
    @staticmethod
    def open_file(path, filename, report_folder=None):
        path = f'{report_folder or REPORT_FOLDER}/{path}'
        open_mode = 'a'
        if not os.path.exists(path):
            os.makedirs(path)

        file_path = f'{path}/{filename}'
        file_exists = True
        if not os.path.isfile(file_path):
            open_mode = 'w+'
            file_exists = False

        return file_exists, open(file_path, open_mode, newline='')

    @staticmethod
    def dict_to_csv(path, filename, dict_list, headers):
        file_exists, f = CSVWriter.open_file(path, filename)
        with f:
            writer = csv.DictWriter(f, fieldnames=headers)
            if not file_exists:
                writer.writeheader()
                for d in dict_list:
                    writer.writerow(d)



import logging

from estrade.exceptions import StopLimitException

logger = logging.getLogger(__name__)


class StopLimitMixin:
    def __init__(self, type_, trade, value):
        """
        Init new stop/limit
        :param type_: <str> in ['STOP', 'LIMIT']
        :param trade: <estrade.trade.Trade> instance
        :param value: <int>
        """
        self.type_ = type_
        self.trade = trade
        self.value = value
        logger.debug('Init new stop/limit: %s' % self.__repr__())

    ##################################################
    # TRADE
    ##################################################
    @property
    def trade(self):
        """
        return trade
        :return: <estrade.trade.Trade> instance
        """
        return self._trade

    @trade.setter
    def trade(self, trade):
        """
        Set trade
        :param trade: <estrade.trade.Trade> instance
        :return:
        """
        from estrade.trade import Trade

        if not isinstance(trade, Trade):
            raise StopLimitException('Invalid Trade on Stop/Limit')
        self._trade = trade

    ##################################################
    # TYPE
    ##################################################
    @property
    def type_(self):
        """
        return stop/limit type
        :return:<str>
        """
        return self._type_

    @type_.setter
    def type_(self, type_):
        """
        Set stop/limit type
        :param type_: <str> in ['STOP', 'LIMIT']
        :return:
        """
        if not isinstance(type_, str) or type_.upper() not in ['STOP', 'LIMIT']:
            raise StopLimitException(f'Invalid type {type_}')
        self._type_ = type_.upper()

    ##################################################
    # VALUES
    ##################################################
    @property
    def value(self):
        """
        return absolute value of stop/limit (value provided on init)
        :return: <int>
        """
        raise StopLimitException(
            'Cannot call value for stop/limit, call absolute_value '
            'or relative_value explicitly.'
        )

    @value.setter
    def value(self, value):
        """
        Check value type is int/float and check consistency
        :param value: <int> or <float>
        :return:
        """

        if type(value) not in [int, float]:
            raise StopLimitException('Stop/limit cannot be a non-numeric value')

        if value <= 0:
            raise StopLimitException('Stop/limit cannot be negative')

        self._check_value_consistency(value)
        self._value = value

        logger.debug(repr(self))
        # TODO: update relative value on trade open update

    @property
    def absolute_value(self):
        """
        This method returns absolute value of stop/limit.
        :return: <int> or <float>
        """
        raise NotImplementedError()

    @property
    def relative_value(self):
        """
        This method returns absolute value of stop/limit.
        :return: <int> or <float>
        """
        raise NotImplementedError()

    def _check_value_consistency(self, value):
        """
        This method must be implemented to check that the provided stop/limit value is
        consistent with trade open.
        :param value: <int> or <float>
        :return:
        """
        raise NotImplementedError()

    ##################################################
    # DIRECTION
    ##################################################
    @property
    def direction(self):
        """
        Depending of stop/limit type, the direction usage must be implemented
        :return:
        """
        raise NotImplementedError()

    @property
    def is_relative(self):
        """
        Check if stop/limit is relative
        :return: <bool>
        """
        return False

    @property
    def is_absolute(self):
        """
        Check if stop/limit is absolute
        :return: <bool>
        """
        return False

    ##################################################
    # CHECK
    ##################################################
    def check(self, tick):
        """
        Check stop/limit value against a tick:
            - returns True if stop/limit reached
            - returns False if stop/limit not reached
        :param tick: <estrade.tick.Tick> instance
        :return: <bool>
        """
        if self.type_ == 'STOP':
            if self.trade.direction > 0 and tick.bid <= self.absolute_value:
                logger.debug(
                    'close BUY because tick %f  <= Stop absolute value %f'
                    % (tick.bid, self.absolute_value)
                )
                return True
            elif self.trade.direction < 0 and tick.ask >= self.absolute_value:
                logger.debug(
                    'close SELL because tick %f >= Stop absolute value %f'
                    % (tick.ask, self.absolute_value)
                )
                return True
        elif self.type_ == 'LIMIT':
            if self.trade.direction > 0 and tick.bid >= self.absolute_value:
                logger.debug(
                    'close BUY because tick %f >= limit absolute value %f'
                    % (tick.bid, self.absolute_value)
                )
                return True
            elif self.trade.direction < 0 and tick.ask <= self.absolute_value:
                logger.debug(
                    'close SELL because tick %f <= limit absolute value %f'
                    % (tick.ask, self.absolute_value)
                )
                return True
        else:
            logger.debug('Stop/Limit not reached.')

        return False

    def __repr__(self):
        """
        Representation of a stop/limit
        :return: <dict>
        """
        repr = f'{self.type_} : direction: {self.direction} of'
        if self.is_absolute:
            return f'absolute {repr} {self.absolute_value}'
        elif self.is_relative:
            return f'relative {repr} {self.relative_value}'
        else:
            return 'undefined stop/limit'



__version__ = '0.0.1'

# classes
from estrade.strategy import Strategy  # noqa
from estrade.mixins.reporting_mixin import ReportingMixin  # noqa

from estrade.candle import Candle  # noqa
from estrade.candle_set import CandleSet  # noqa
from estrade.epic import Epic  # noqa
from estrade.market import Market  # noqa
from estrade.tick import Tick  # noqa
from estrade.provider import Provider  # noqa

# indicators
from estrade.indicators.candle_set.moving_average import (  # noqa
    ExponentialMovingAverage,  # noqa
    SimpleMovingAverage,  # noqa
)  # noqa

# reporting
from estrade.reporting.csv import ReportingCSV  # noqa

# logging
import logging

logging.basicConfig(level=logging.DEBUG)
logging.getLogger(__name__).addHandler(logging.NullHandler())



""" This module define the Trade class used to manage positions.
"""
import logging

from estrade.mixins.ref_mixin import RefMixin
from estrade.exceptions import TradeException
from estrade.observer import Observable
from estrade.stop_limit import StopLimitMixinAbsolute, StopLimitMixinRelative
from estrade.tick import Tick


logger = logging.getLogger(__name__)


class TradeClose:
    """
    Manage a trade close
    """

    def __init__(self, tick, quantity, result, reason):
        """
        Init a trade close
        :param tick: <estrade.tick.Tick> instance
        :param quantity: <int>
        :param result: <float>
        :param reason: <str>
        """
        self.tick = tick
        self.quantity = quantity
        self.result = result
        self.reason = reason

    def to_json(self):
        """
        Convert trade close to dict (mainly used for reporting)
        :return: <dict>
        """
        return {
            'tick': self.tick.to_json(True),
            'quantity': self.quantity,
            'reason': self.reason,
            'result': self.result,
        }


class Trade(RefMixin, Observable):
    """
        The Trade class manage trades (=market positions)
    """

    def __init__(
        self,
        quantity,
        direction,
        tick=None,
        epic=None,
        trade_manager=None,
        strategy=None,
        ref=None,
        stop_absolute=None,
        stop_relative=None,
        limit_absolute=None,
        limit_relative=None,
        meta=None,
    ):
        """
        Init a new trade
        :param tick: <estrade.tick.Tick> instance : trade open tick
        :param epic: <str> epic ref
        :param quantity: <int> opened quantity
        :param trade_manager: <estrade.trade_manager.TradeManager> trade manager
        managing this trade
        :param direction: <int> in [1, -1, 'BUY', 'SELL'] trade direction (buy or sell)
        :param strategy: <estrade.Astrategy.Astrategy> child instance => strategy that
        generated this trade
        :param ref: <str> trade reference
        :param stop_absolute: <int> absolute stop
        :param stop_relative: <int> relative stop
        :param limit_absolute: <int> absolute limit
        :param limit_relative: <int> relative limit
        :param meta: <dict> free of use dict to register provider events etc.
        """
        # set ref with parent class
        RefMixin.__init__(self, ref)
        # set class as observable so it can fire events
        # TODO : fire events on trade update
        Observable.__init__(self)

        self.trade_manager = trade_manager
        self.strategy = strategy
        self.direction = direction
        self.inital_quantities = abs(quantity)
        self.opened_quantity = abs(quantity)
        self.closed_quantity = 0
        self.closed = False

        # init empty list of ticks received by trade
        self.ticks = []

        self._add_new_tick(self._get_open_tick(tick, epic))

        # init empty list of closes
        self.closes = []

        # set max gain and max losses
        self.max_gain = float('inf') * -1
        self.max_loss = float('inf')
        self.update_max_loss_gains(self._calculate_result())

        # init stop/limit from params
        self._stop = None
        if stop_relative:
            self.set_stop(stop_value=stop_relative, relative=True)
        elif stop_absolute:
            self.set_stop(stop_value=stop_absolute, relative=False)

        self._limit = None
        if limit_relative:
            self.set_limit(limit_value=limit_relative, relative=True)
        elif limit_absolute:
            self.set_limit(limit_value=limit_absolute, relative=False)

        self.meta = {} if not meta or not isinstance(meta, dict) else meta

        logger.info('%s' % repr(self))

    ##################################################
    # INIT HELPERS
    ##################################################
    def _get_open_tick(self, tick, epic):
        # add tick to list of ticks
        if tick is not None:
            logger.warning(
                'It is highly disrecommended to specify tick on trade open. '
                'A trade whould be opened with not specified tick '
                '(the last epic tick will be used)'
            )
            # check that provided tick is valid
            Trade.is_open_tick_valid(tick)
            return tick
        elif epic is not None:
            epic = self.trade_manager.market.get_epic(epic)
            if not epic:
                raise TradeException(
                    'Impossible to find epic (to get the trade open tick)'
                )
            return epic.current_tick

        raise TradeException('Provide epic ref to open trade')

    @staticmethod
    def is_open_tick_valid(tick):
        """
        Check that tick provided is valid
        :param tick: <estrade.tick.Tick> instance
        :return: <bool> True if tick valid and epic tradeable else False
        """
        if not Trade.is_tick_valid(tick):
            raise TradeException('Invalid tick')

        if not Trade.is_tick_tradeable(tick):
            raise TradeException(
                f'Tick epic {tick.epic.ref} is not tradeable, cannot open trade'
            )

        return True

    @staticmethod
    def is_tick_valid(tick):
        """
        Check if a tick is a valid class instance
        :param tick: <estrade.tick.Tick> instance
        :return: <bool>
        """
        return isinstance(tick, Tick)

    @staticmethod
    def is_tick_tradeable(tick):
        """
        Check if tick epic is tradeable
        :param tick: <estrade.tick.Tick> instance
        :return: <bool>
        """
        return tick.epic.tradeable

    ##################################################
    # TRADE MANAGER GET/SET
    ##################################################
    @property
    def trade_manager(self):
        """
        return trade manager that holds this trade
        :return: <estrade.trade_manager.TradeManager>
        """
        return self._trade_manager

    @trade_manager.setter
    def trade_manager(self, trade_manager):
        """
        Set the trade manager that holds this trade
        :param trade_manager: <estrade.trade_manager.TradeManager>
        :return:
        """
        # import here to prevent import loop
        from estrade.trade_manager import TradeManager

        if trade_manager is None:
            self._trade_manager = None
            return

        if not isinstance(trade_manager, TradeManager):
            raise TradeException(f'Invalid trade manager: {trade_manager}')
        self._trade_manager = trade_manager

    ##################################################
    # DIRECTION GET/SET
    ##################################################
    @property
    def direction(self):
        """
        Return trade direction
        :return: <int> 1 if trade is a BUY, -1 if trade is a SELL
        """
        return self._direction

    @direction.setter
    def direction(self, direction):
        """
        Set trade direction
        :param direction: <int> in [-1, 1] or <str> in ['SELL', 'BUY']
        :return:
        """
        if direction:
            if direction == 1 or (
                isinstance(direction, str) and direction.upper() == 'BUY'
            ):
                self._direction = 1
                return
            elif direction == -1 or (
                isinstance(direction, str) and direction.upper() == 'SELL'
            ):
                self._direction = -1
                return
        raise TradeException(f'Invalid trade direction: {direction}')

    ##################################################
    # Strategy GET/SET
    ##################################################
    @property
    def strategy(self):
        """
        Returns the strategy that generated this trade
        :return: <estrade.strategy.Strategy> child instance or None
        """
        return self._strategy

    @strategy.setter
    def strategy(self, strategy):
        """
        Set trade strategy:
            - if strategy is provided in params, set this strategy
            - else if only one strategy is registered on market -> use this strategy
            - else raise error
        :param strategy: <estrade.strategy.Strategy> child instance or None
        :return:
        """
        from estrade.strategy import Strategy  # to prevent import loop

        if strategy is None:
            if (
                self.trade_manager
                and self.trade_manager.market
                and len(self.trade_manager.market.strategies) == 1
            ):
                logger.warning(
                    'Use market unique strategy to open trade, \
                               it is recommended to explicitly define \
                               strategy when opening trade'
                )
                self._strategy = self.trade_manager.market.strategies[0]
                return
            self._strategy = None
            logger.warning('A trade was opened without a strategy defined')
            return

        if not isinstance(strategy, Strategy):
            raise TradeException(f'Invalid Strategy : {strategy}')

        self._strategy = strategy

    ##################################################
    # STOP/LIMIT HELPERS
    ##################################################
    def _set_stop_limit(self, type_, value, relative):
        """
        Set a stop or a limit on trade
        :param type_: <str> in ['STOP', 'LIMIT']
        :param value: <int> or <float>
        :param relative: <bool>
        :return: <estrade.stop_limit.AbstractStopLimit> child instance
        """
        if value is None:
            return None

        if relative:
            # set stop/limit as relative
            return StopLimitMixinRelative(type_=type_, trade=self, value=value,)

        # set stop/limit as an absolute stop
        return StopLimitMixinAbsolute(type_=type_.lower(), trade=self, value=value)

    ##################################################
    # STOP GET/SET
    ##################################################
    @property
    def stop(self):
        """
        Return trade stop
        :return: <estrade.stop_limit.AbstractStopLimit> Child
        """
        return self._stop

    @stop.setter
    def stop(self, stop):
        """
        Prevent set of stop directly as stop set requires to define if the stop is
        relative (use set_stop method)
        :param stop:
        :return:
        """
        raise NotImplementedError(
            'Impossible to directly set stop on trade, use Trade.set_stop method'
        )

    def set_stop(self, stop_value, relative=False):
        """
        Stop setter
        :param stop_value: <int> or <float>
        :param relative: <bool> : is the stop value provided is relative to trade
        open tick?
        :return:
        """
        self._stop = self._set_stop_limit(
            type_='stop', value=stop_value, relative=relative
        )

    ##################################################
    # LIMIT GET/SET
    ##################################################
    @property
    def limit(self):
        """
        return trade limit
        :return: <estrade.stop_limit.AbstractStopLimit> Child
        """
        return self._limit

    @limit.setter
    def limit(self, limit):
        """
        Prevent set of limit directly as limit set requires to define if the limit is
        relative (use set_limit method)
        :param limit:
        :return:
        """
        raise NotImplementedError(
            'Impossible to directly set limit on trade, use Trade.set_limit method'
        )

    def set_limit(self, limit_value, relative=False):
        """
        Limit setter
        :param limit_value: <int> or <float>
        :param relative: <bool> : is the stop value provided is relative to trade open
        tick?
        :return:
        """
        self._limit = self._set_stop_limit(
            type_='limit', value=limit_value, relative=relative
        )

    ##################################################
    # EPIC GETTERS
    ##################################################
    @property
    def epic(self):
        """
        get trade epic
        :return: <estrade.epic.Epic> instance
        """
        return self.ticks[0].epic

    ##################################################
    # OPEN GETTERS
    ##################################################
    @property
    def open_tick(self):
        """
        get trade open tick
        :return: <estrade.tick.Tick> instance
        """
        return self.ticks[0]

    @property
    def last_tick(self):
        return self.ticks[-1] if self.ticks else None

    @property
    def open(self):
        """
        Return trade open value (value of open tick)
        :return: <int>
        """
        if self.direction > 0:
            return self.open_tick.ask

        return self.open_tick.bid

    ##################################################
    # MIN/MAX GETTERS
    ##################################################
    @property
    def max(self):
        """
        Return upper value of ticks received during this trade
        # TODO: rename this property for clarity
        :return: <float>
        """
        return max([t.value for t in self.ticks])

    @property
    def min(self):
        """
        Return lower value of ticks received during this trade
        # TODO: rename this property for clarity
        :return: <float>
        """
        return min([t.value for t in self.ticks])

    ##################################################
    # QUANTITIES GETTERS
    ##################################################
    @property
    def quantity(self):
        """
        Return current opened quantities
        :return: <int>
        """
        return self.opened_quantity

    ##################################################
    # CLOSE
    ##################################################
    def close(self, tick=None, quantity=None, close_reason='manual'):
        """
        Total or partial close of trade.

        :param tick: <estrade.tick.Tick> instance -> tick to use for close
            (if not provided, last received tick will be used)
        :param quantity: <int> -> quantities to close
            (if not provided, all opened quantities will be closed)
        :param close_reason: <str> -> comment on trade close
        :return:
        """
        # define quantities to close
        if quantity and quantity > self.opened_quantity:
            raise TradeException(
                f'Impossible to close {quantity} '
                f'(more than remaining quantity {self.opened_quantity}'
            )
        if not quantity:
            quantity = self.opened_quantity

        # define tick to use for close
        if not tick:
            tick = self.ticks[-1]
        elif not Trade.is_tick_valid(tick):
            raise TradeException('Invalid tick provided to close trade')

        # prevent close when epic is not tradeable
        if not Trade.is_tick_tradeable(tick):
            logger.error(
                'Impossible to close trade because epic %s is not tradeable.'
                % tick.epic.ref
            )
            return

        # calculate result of this close
        close_result = self._calculate_result(close_tick=tick, quantity=quantity)

        # add a new close object to list of closed.
        self.closes.append(
            TradeClose(
                tick=tick, quantity=quantity, result=close_result, reason=close_reason,
            )
        )
        self.closed_quantity += quantity
        self.opened_quantity -= quantity
        if self.opened_quantity <= 0:
            self.closed = True

        logger.info(
            '%s : Close %d @ %f > %f (max loss: %f, max_gain: %f)'
            % (
                tick.datetime,
                quantity,
                tick.value,
                close_result,
                self.max_loss,
                self.max_gain,
            )
        )

    ##################################################
    # RESULT
    ##################################################
    def _calculate_result(self, close_tick=None, quantity=None):
        """
        Calculate trade result for close.
        :param sell_tick: <estrade.tick.Tick> instance -> tick to use to calculate
        close result
            (if not provided, the last tick received by trade will be used)
        :param quantity: <int> -> quantities to close
        :return:
        """
        if not close_tick:
            close_tick = self.ticks[-1]
        if not quantity:
            quantity = self.opened_quantity

        if self.direction > 0:
            return round((close_tick.bid - self.ticks[0].ask) * abs(quantity), 2)
        else:
            return round((self.ticks[0].bid - close_tick.ask) * abs(quantity), 2)

    @property
    def closed_result(self):
        """
        Result of closed quantities for this trade
        :return: <float>
        """
        result = 0
        for close in self.closes:
            result += close.result
        return result

    @property
    def opened_result(self):
        """
        Current result of opened quantities
        :return:
        """
        return self._calculate_result()

    @property
    def result(self):
        """
        Total trade result (opened quantities + closed quantities)
        :return:
        """
        return self.closed_result + self.opened_result

    ##################################################
    # HANDLE NEW TICK
    ##################################################
    def on_new_tick(self, tick):
        """
        Method triggered when a new tick is received by trade_manager.
        :param tick: <estrade.tick.Tick> instance
        :return:
        """
        if not self.closed:
            self._add_new_tick(tick)

            # update max gain and max loss
            self.update_max_loss_gains(self._calculate_result())

            # check stop/limit
            logger.debug(
                'Check if tick value reached stop (%s) or limit (%s)'
                % (self.stop, self.limit)
            )
            if self.stop and self.stop.check(tick):
                if self.trade_manager:
                    self.trade_manager.close_trade(trade=self, close_reason='stop')
                else:
                    self.close(close_reason='stop')
            elif self.limit and self.limit.check(tick):
                if self.trade_manager:
                    self.trade_manager.close_trade(trade=self, close_reason='limit')
                else:
                    self.close(close_reason='limit')

    def _add_new_tick(self, tick):
        """
        Method to add a new tick to trade
        :param tick: <estrade.tick.Tick> instance
        :return:
        """
        if not Trade.is_tick_valid(tick):
            raise TradeException('Invalid tick provided to add on trade.')

        self.ticks.append(tick)

    def update_max_loss_gains(self, current_result):
        """
        Update trade max gain and max losses.
        :param current_result: <float>
        :return:
        """
        if current_result < self.max_loss:
            self.max_loss = self.result
        elif current_result > self.max_gain:
            self.max_gain = self.result

    ##################################################
    # REPORTING
    ##################################################
    @property
    def to_json(self):
        return {
            'ref': self.ref,
            'open_at': self.open_tick.datetime,
            'open': self.open_tick.value,
            'open_quantity': abs(self.inital_quantities),
            'direction': self.direction,
            'stop_rel': self.stop.relative_value if self.stop else None,
            'stop_abs': self.stop.absolute_value if self.stop else None,
            'limit_rel': self.limit.relative_value if self.limit else None,
            'limit_abs': self.limit.absolute_value if self.limit else None,
            'closed_quantity': abs(self.closed_quantity),
            'closed': self.closed,
            'closes': [c.to_json() for c in self.closes],
            'result': self.result,
            'max_loss': self.max_loss,
            'max_gain': self.max_gain,
            'nb_ticks': len(self.ticks),
            **self.meta,
        }

    @property
    def json_headers(self):
        base_headers = [
            'ref',
            'open_at',
            'open',
            'open_quantity',
            'direction',
            'stop_rel',
            'stop_abs',
            'limit_rel',
            'limit_abs',
            'closed_quantity',
            'closed',
            'closes',
            'result',
            'max_loss',
            'max_gain',
            'nb_ticks',
        ]
        base_headers.extend(list(self.meta.keys()))
        return base_headers

    def __repr__(self):
        repr_string = (
            f'{self.open_tick.datetime}: '
            f'Trade for {self.inital_quantities} @ {self.open} '
            f'with ref: {self.ref} '
            f'on strategy {self.strategy.ref if self.strategy else None}'
        )
        if self.stop:
            repr_string = (
                f'{repr_string}, '
                f'stop absolute: {self.stop.absolute_value}, '
                f'stop relative {self.stop.relative_value}'
            )
        if self.limit:
            repr_string = (
                f'{repr_string}, '
                f'limit absolute: {self.limit.absolute_value}, '
                f'limit relative {self.limit.relative_value}'
            )
        return repr_string



import logging

from estrade.exceptions import TradeManagerException
from estrade.observer import Observable
from estrade.trade import Trade

logger = logging.getLogger(__name__)


class TradeManager(Observable):
    """
    A trade manager handle a list of trades.

    :param estrade.Market market: market instance
    :param list trades: list of pre-existing trades
    """

    def __init__(self, market=None, trades=None):
        """
        Init a new trade manager
        """
        # init list of trades
        self.trades = trades
        self.open_trades = []
        self.closed_trades = []
        self.strategy_trades = {}

        # init market
        self.market = market

        # set trade manager as observable so it can fire events
        Observable.__init__(self)
        logger.debug(
            'New trade manager created with %d open trades and %d closed trades'
            % (len(self.open_trades), len(self.closed_trades))
        )
        logger.debug('strategy trades: %s' % self.strategy_trades)

    ##################################################
    # Market
    ##################################################
    def add_strategies(self):
        logger.debug('add strategies to trade manager')
        for strategy in self.market.strategies:
            if strategy.ref not in self.strategy_trades:
                logger.debug('add strategy %s to trade manager' % strategy.ref)
                self.strategy_trades[strategy.ref] = {'opened': [], 'closed': []}

    ##################################################
    # Trades
    ##################################################
    @property
    def trades(self):
        """
        Retur list of trades
        :return: [<self.trade_class>,]
        """
        return self._trades

    @trades.setter
    def trades(self, trades):
        """
        Set list of trades
        :param trades: None or [<self.trade_class>,]
        :return:
        """
        self._trades = []
        if not trades:
            return

        if not isinstance(trades, list):
            raise TradeManagerException('Trades must be a list')

        for trade in trades:
            if not isinstance(trade, Trade):
                raise TradeManagerException(
                    'Trades list must only contains instance of Trades'
                )
            self._trades.append(trade)

    def get_trade_by_ref(self, ref):
        """
        Find a trade by its ref
        :param ref: <str>
        :return: <self.trade_class> instance or None
        """
        for trade in reversed(self.trades):
            if trade.ref == ref:
                return trade
        return None

    ##################################################
    # OPEN TRADE
    ##################################################
    def open_trade(self, *args, **kwargs):
        """
        This method create a new trade.

        :param args/kwargs: arguments to instanciate a new instance of the current
        `trade_class` (by default a :class:`estrade.trade.Trade` instance)
        :return: `trade_class` instance
        """
        logger.debug('Open new trade')
        self.fire('trade_manager_before_new_trade', *args, **kwargs)

        new_trade = Trade(*args, trade_manager=self, **kwargs)
        new_trade = self.market.provider.open_trade(new_trade)
        if not new_trade:
            logger.error(
                'New trade creation failed, do not add it to trade manager trades.'
            )
            return None

        self.trades.append(new_trade)
        self.open_trades.append(new_trade)
        if new_trade.strategy:
            self.strategy_trades[new_trade.strategy.ref]['opened'].append(new_trade)

        self.fire('trade_manager_after_open_trade', trade=new_trade)

        return new_trade

    @property
    def open_buys(self):
        """
        List all opened buy trades
        :return: [<self.trade_class>,]
        """
        return [trade for trade in self.open_trades if trade.direction > 0]

    @property
    def open_sells(self):
        """
        List all opened sell trades
        :return: [<self.trade_class>,]
        """
        return [trade for trade in self.open_trades if trade.direction < 0]

    ##################################################
    # ON NEW TICK EVENT
    ##################################################
    def on_new_tick(self, tick):
        """
        On new tick, dispatch tick to all opened trades
        :param tick: <estrade.tick.Tick>
        :return:
        """
        for trade in reversed(self.open_trades):
            if tick.epic == trade.epic:
                trade.on_new_tick(tick)

    def on_new_ticks_high_low(self, epic, high, low):
        for trade in reversed(self.open_trades):
            if trade.epic == epic:
                if trade.direction > 0:
                    trade.on_new_tick(low)
                    trade.on_new_tick(high)
                else:
                    trade.on_new_tick(high)
                    trade.on_new_tick(low)

    ##################################################
    # CLOSE TRADE
    ##################################################
    def close_trade_api_call(self, trade, quantity):
        """
        Method used by providers trade manager to close trade by api.
        :param trade: <self.trade_class> instance
        :param quantity: <int>
        :return: <self.trade_class> instance or None if api call failed
        """

        return trade

    def close_trade(
        self, trade, quantity=None, close_reason='close unique trade', tick=None
    ):
        """
        method to close a trade.

        This method fire the following events:
            - trade_manager_before_close_trade
            - trade_manager_after_close_trade

        :param trade: <self.trade_class> instance
        :param quantity: <int> quantity to close on trade
        :param close_reason: <str> closing reason
        :param tick: <estrade.tick.Tick> instance : tick to use for trade close
        :return:
        """
        logger.info('close unique trade %s', trade)
        quantity = (
            quantity
            if quantity and quantity <= trade.opened_quantity
            else trade.opened_quantity
        )

        self.fire(
            'trade_manager_before_close_trade',
            trade=trade,
            quantity=quantity,
            close_reason=close_reason,
        )

        updated_trade = self.market.provider.close_trade(trade, quantity)
        if not updated_trade:
            logger.error(
                'Impossible to close %d on trade because api call failed for %s'
                % (quantity, trade.ref)
            )
            return None

        updated_trade.close(quantity=quantity, close_reason=close_reason, tick=tick)

        # drop trade from open_trades and add it to closed trades
        self.open_trades.remove(updated_trade)
        self.closed_trades.append(updated_trade)

        if updated_trade.strategy:
            self.strategy_trades[updated_trade.strategy.ref]['opened'].remove(
                updated_trade
            )
            self.strategy_trades[updated_trade.strategy.ref]['closed'].append(
                updated_trade
            )

        self.fire('trade_manager_after_close_trade', trade=updated_trade)

    def close_trade_by_ref(
        self,
        ref,
        close_reason='close trade by ref',
        quantity=None,
        strategy=None,
        tick=None,
    ):
        """
        Close a trade by its reference
        :param ref: <str> ref of a trade in self.trades
        :param close_reason: <str>
        :param quantity: <int> quantity to close
        :param strategy: <estrade.strategy.Strategy> strategy to restrict search for
        trade on
        :param tick: <estrade.tick.Tick> instance : tick to use for trade close
        :return:
        """
        for trade in reversed(self.get_trades(only_opened=True, strategy=strategy)):
            if (
                not trade.closed
                and trade.ref == ref
                and (not strategy or trade.strategy == strategy)
            ):
                self.close_trade(
                    trade, quantity=quantity, close_reason=close_reason, tick=tick
                )
                return
        raise TradeManagerException(
            f'impossible to find trade to close with ref {ref} '
            f'in {[trade.ref for trade in self.trades if not trade.closed]}'
        )

    def close_all_trades(
        self, close_reason='close all trades', strategy=None, tick=None
    ):
        """
        Close all open trades (for a specific strategy if prrovided)
        :param close_reason: <str>
        :param strategy: <estrade.strategy.Strategy> child instance
        :param tick: <estrade.tick.Tick> instance : tick to use for trade close
        :return:
        """
        logger.info(
            'close all (%d) trades for strategy %s'
            % (len(self.open_trades), strategy.ref if strategy else None)
        )
        for trade in self.open_trades[:]:
            if not strategy or trade.strategy == strategy:
                self.close_trade(trade, close_reason=close_reason, tick=tick)
                logger.debug('after close trade %s' % trade)
            else:
                logger.debug('do not close trade %s' % trade)

        logger.debug('%d trades remaining in open_trades' % len(self.open_trades))

    def close_all_buys(self, close_reason='close all buys', strategy=None, tick=None):
        """
        Close all opened buy trades (for a specific strategy if provided)
        :param close_reason: <str>
        :param strategy: <estrade.strategy.Strategy> child instance
        :param tick: <estrade.tick.Tick> instance : tick to use for trade close
        :return:
        """
        for trade in self.open_trades[:]:
            if (
                not trade.closed
                and trade.direction > 0
                and (not strategy or trade.strategy == strategy)
            ):
                self.close_trade(trade, close_reason=close_reason, tick=tick)

    def close_all_sells(self, close_reason='close all sells', strategy=None, tick=None):
        """
        Close all opened sell trades (for a specific strategy if provided)
        :param close_reason: <str>
        :param strategy: <estrade.strategy.Strategy> child instance
        :param tick: <estrade.tick.Tick> instance : tick to use for trade close
        :return:
        """
        for trade in self.open_trades[:]:
            if (
                not trade.closed
                and trade.direction < 0
                and (not strategy or trade.strategy == strategy)
            ):
                self.close_trade(trade, close_reason=close_reason, tick=tick)

    ##################################################
    # STATS
    ##################################################
    def get_trades(self, only_opened=False, only_closed=False, strategy=None):
        """
        Get a list of trade based on search input params

        :param only_opened: (bool) only return opened trades
        :param only_closed: (bool) only return closed trades
        :param :class:`estrade.Strategy` strategy: Strategy instance
        :return: [:class:`estrade.trade.Trade`,] list of trades
        """
        if strategy:
            if only_closed:
                return self.strategy_trades[strategy.ref]['closed']
            elif only_opened:
                return self.strategy_trades[strategy.ref]['opened']
            else:
                return (
                    self.strategy_trades[strategy.ref]['opened']
                    + self.strategy_trades[strategy.ref]['closed']
                )

        if only_closed:
            return self.closed_trades
        elif only_opened:
            return self.open_trades
        return self.trades

    def won_trades(self, **kwargs):
        """
        Return list of trade with a positive (or zero) result
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: [<self.trade_class>,]
        """
        return [trade for trade in self.get_trades(**kwargs) if trade.result >= 0]

    def lost_trades(self, **kwargs):
        """
        Return list of trade with a negative result
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: [<self.trade_class>,]
        """
        return [trade for trade in self.get_trades(**kwargs) if trade.result < 0]

    def nb_trades(self, **kwargs):
        """
        Count number of trade matching input argements
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: [<self.trade_class>,]
        """
        return len(self.get_trades(**kwargs))

    def nb_win(self, **kwargs):
        """
        Count number of trade with a positive result and matching input parameters
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: <int>
        """
        return len(self.won_trades(**kwargs))

    def nb_losses(self, **kwargs):
        """
        Count number of trade with a negative result and matching input parameters
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: <int>
        """
        return len(self.lost_trades(**kwargs))

    def total_win(self, **kwargs):
        """
        Return the total result of trades with a positive result
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: <int>
        """
        return sum([trade.result for trade in self.won_trades(**kwargs)])

    def total_losses(self, **kwargs):
        """
        Return the total result of trades with a negative result
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: <int>
        """
        return sum([trade.result for trade in self.lost_trades(**kwargs)])

    def result(self, **kwargs):
        """
        Return the total result of trades matching input arguments
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: <int>
        """
        return sum([trade.result for trade in self.get_trades(**kwargs)])

    def profit_factor(self, **kwargs):
        """
        Return the profit factor of trades matching input arguments

        Note:
            profit factor is absolute value of (
                sum of result of trades with positive result /
                sum of results of trades with negative results
            )
        :param kwargs: (same arguments as TradeManager.get_trades method)
        :return: <float>
        """
        total_win = self.total_win(**kwargs)
        total_losses = self.total_losses(**kwargs)

        if total_losses == 0:
            return total_win
        return total_win / abs(total_losses)

    def max_drowdown(self, **kwargs):
        """
        return the max drowdown of a list of trades
        :param kwargs:
        :return:
        """
        max_dd = 0
        min = 0
        max = 0
        result = 0
        for trade in self.get_trades(**kwargs):
            result += trade.result
            if result > max:
                max = result
                min = result
            elif result < min:
                min = max - result
                if min > max_dd:
                    max_dd = min
        return max_dd



import logging

from estrade.mixins.reporting_mixin import ReportingMixin
from estrade.utils.csv import CSVWriter

logger = logging.getLogger(__name__)


class ReportingCSV(ReportingMixin):
    @property
    def base_path(self):
        return self.market.ref

    def strategy_path(self, strategy):
        return f'{self.market.ref}/{strategy.ref}/'

    def on_new_tick(self, tick):
        pass

    def on_trade_update(self, trade):
        pass

    def on_run_end(self):
        logger.info('Report as CSV')
        strategies_dicts = []
        for strategy in self.market.strategies:
            strategies_dicts.append(strategy.to_json)

            strategy_trades_dict = []
            headers = []
            for trade in self.market.trade_manager.get_trades(strategy=strategy):
                strategy_trades_dict.append(trade.to_json)
                if not headers:
                    headers = trade.json_headers

            consolidated_report_file = 'consolidated_trades.csv'
            CSVWriter.dict_to_csv(
                path=self.strategy_path(strategy),
                filename=consolidated_report_file,
                dict_list=strategy_trades_dict,
                headers=headers,
            )

        CSVWriter.dict_to_csv(
            path=self.base_path,
            filename='strategies_report.csv',
            dict_list=strategies_dicts,
            headers=self.market.strategies[0].json_headers,
        )
        logger.info('Reporting file was writen in %s' % self.base_path)
        print(strategies_dicts)



import logging
from typing import Any, Dict, List, Optional, TYPE_CHECKING

import arrow

from estrade.candle_set import CandleSet
from estrade.candle import Candle
from estrade.mixins.ref_mixin import RefMixin
from estrade.mixins.candle_set_indicator_mixin import CandleSetIndicatorMixin
from estrade.epic import Epic
from estrade.tick import Tick
from estrade.trade import Trade
from estrade.exceptions import StrategyException

if TYPE_CHECKING:
    from estrade.market import Market

logger = logging.getLogger(__name__)


class Strategy(RefMixin):
    """
    This class define a strategy to apply on ticks/candles.t pu

    Arguments:
        epics: list of `estrade.Epic` instances
        meta: free of use dict of parameters you can re-use in your strategy methods
        ref: strategy name
        max_concurrent_trades: nb max of trades concurrently opened

    Define your custom strategies to define when to open/close trades. Your strategies
    are triggered by all new ticks received and all new candles created on the
    Epics attached to it.

    !!! example
        Example of a strategy that

         - open a BUY trade when the tick value is 1000
         - open a SELL trade when the last closed candle was red
         - close all opened BUY trades when the moving average on 50 periods is
         inferior to the current tick - 20
         - close all trades when the strategy result is > 200 when a new candle opens.

        ```python
        from estrade import CandleSet, Market, Provider, SimpleMovingAverage, Strategy

        class MyCustomStrategy(Strategy):
            def on_new_tick_opening_strategy(tick):
                if tick.value == 1000:
                    self.open_trade(
                        epic=tick.epic.ref,
                        quantity=2,
                        direction='SELL',
                        stop_absolute=1010,
                        limit_absolute=990,
                    )

            def on_new_candle_opening_strategy(candle_set):
                last_candle = candle_set.last_closed_candle
                # check if is defined because on first candle in candle set,
                # it will be none.
                if last_candle:
                    if last_candle.color == 'red':
                        self.open_trade(
                            epic=candle_set.epic.ref,
                            quantity=1,
                            direction='BUY',
                            stop_relative=30,
                        )

            def on_new_tick_closing_strategy(tick):
                mm50_15mn = self.get_indicator(
                    epic_ref='MY_EPIC',
                    timeframe='15minutes',
                    indicator_name='mm50'
                )
                # check if exists because for the 49 fist candles it will not be defined
                if mm50_15mn:
                    if mm50_15mn < (tick.value - 20):
                        self.close_all_buys()


            def on_new_candle_closing_strategy(tick):
                result = self.market.trade_manager.result(strategy=self)
                if result > 200:
                    self.close_all_trades()

        class MyProvider(Provider):
            # define your custom provider
            pass

        if __name__ == '__main__':
            market = Market()

            epic = Epic(ref='MY_EPIC', market=market)
            mm50 = SimpleMovingAverage(ref='mm50')
            candle_set = CandleSet(epic=epic, timeframe='15minutes', indicators=[mm50])

            strategy = MyCustomStrategy(market=market, epics=[epic])
            provider = MyProvider(market=market)

            market.run()

        ```


    """

    def __init__(
        self,
        market: 'Market',
        epics: List[Epic],
        meta: Optional[Dict[str, Any]] = None,
        ref: str = None,
        max_concurrent_trades: int = 1,
    ):
        logger.info('Init new strategy creation')
        self.market: Market = market
        self.market.strategies.append(self)
        # init class ref attribute
        RefMixin.__init__(self, ref)

        self.epics = epics
        self.meta = meta or {}
        self.max_concurrent_trades = max_concurrent_trades
        self.status = "active"

        self.paused_until: Optional[arrow.Arrow] = None
        self.stopped = False

        logger.info('New strategy created %s' % self.ref)

    ##################################################
    # EPICS
    ##################################################
    @property
    def epics(self) -> List[Epic]:
        """
        Return:
            Returns list of epics attached to strategy
        """
        return self._epics

    @epics.setter
    def epics(self, epics: List[Epic]) -> None:
        """
        Attach epics on the current strategy.

        Raises:
            StrategyExeption: if input is not a list of `estrade.Epic` instances.


        """
        self._epics = []

        if not isinstance(epics, list):
            raise StrategyException('Invalid epics on strategy creation')

        for epic in epics:
            if not isinstance(epic, Epic):
                raise StrategyException('Invalid Epic')
            logger.debug('Add epic %s to strategy %s' % (epic.ref, self.ref))
            self._epics.append(epic)
            epic.strategies.append(self)

    ##################################################
    # GETTERS
    # Helpers to fetch elements in strategy execution methods
    ##################################################
    def get_epic(self, epic_ref: str) -> Epic:
        """
        Getter to find epic in strategies epic from its ref.

        Arguments:
            epic_ref: ref of an Epic in the current strategy epics.

        Raises:
            StrategyException: if epic not found in current strategy epics

        Returns:
            Epic instance

        """
        for epic in self.epics:
            if epic.ref == epic_ref:
                return epic
        raise StrategyException(f'Epic with ref {epic_ref} not found')

    def get_candle_set(self, epic_ref: str, timeframe: str) -> CandleSet:
        """
        Getter to find a candle set from epic_ref and timeframe.

        Arguments:
            epic_ref: reference of an epic attached to this strategy
            timeframe: timeframe of the `estrade.CandleSet` to return

        Raises:
            StrategyException : if epic not found on market
            EpicException : if no candle set with this timeframe found on this epic

        Return:
            A candle set object, see `estrade.CandleSet`

        """
        return self.get_epic(epic_ref).get_candle_set(timeframe)

    def get_candle(
        self, epic_ref: str, timeframe: str, offset: int = 0
    ) -> Optional[Candle]:
        """
        Getter to find a candle by its offset in a candleset

        Arguments:
            epic_ref: reference of an epic attached to this strategy
            timeframe: timeframe of the `estrade.CandleSet` to return
            offset: index of the candle to retrieve

        !!! info "Offset usage"
            - 0: returns the currently opened candle on candleset
            - 1 : returns the last closed candle
            - etc.

        !!! tip
            Reserve usage of this getter when you need to find candles that are not the
            current candle or the last closed candle.

            It is recommendend that you call the `Strategy.get_candle_set` method and
            retrieve its `last_closed_candle`, `current_candle` for these cases.

        Raises:
            StrategyException : if epic not found on market
            EpicException : if no candle set with this timeframe found on this epic

        Return:
            A candle object, see `estrade.Candle`

        """
        offset = abs(offset) + 1
        cs = self.get_candle_set(epic_ref, timeframe)
        if len(cs.candles) >= offset:
            return cs.candles[offset * -1]
        return None

    def get_indicator(
        self, epic_ref: str, timeframe: str, indicator_name: str, offset: int = 0
    ) -> Optional[CandleSetIndicatorMixin]:
        """
        Getter to find a candle indicator value.

        Arguments:
            epic_ref: ref of an epic in current strategy epics
            timeframe: matching a candleset.timeframe in epic
            indicator_name: matching an existing indicator on candle_set
            offset: offset of the indicator

        Raise:
            StrategyException : if epic not found on market
            EpicException : if no candle set with this timeframe found on this epic

        Returns:
            Indicator value
        """
        candle = self.get_candle(epic_ref, timeframe, offset)
        if not candle:
            return None
        return candle.indicators.get(indicator_name)

    def get_trades(self, **kwargs) -> Optional[List[Trade]]:
        """
        Getter to find list of strategy trades.

        !!! seealso
            referer to `TradeManager.get_trades` to view the list of arguments

        Returns:
            List of trades attached to the current strategy
        """
        return self.market.trade_manager.get_trades(strategy=self, **kwargs)

    ##################################################
    # OPEN/CLOSE trade(s) helpers
    ##################################################
    def open_trade(self, **kwargs) -> None:
        """
        Use this method to open a trade.

        This method calls `estrade.trade_manager.TradeManager.open_trade` method
        and inject this strategy as an argument.

        As a result, a new `estrade.Trade` instance is created and attached to the
        TradeManager.

        for example, you can open a trade in your strategy event as following ::

        !!! example
            Example of trade open that you can use in opening strategies
            ```python
                from estrade import Strategy

                class MyCustomStrategy:
                    def on_new_tick_opening_strategy(self, tick):
                        if tick.value == 1000:
                            self.open_trade(
                                # mandatory params
                                epic=tick.epic.ref,
                                quantity=5,
                                direction='SELL',
                                # optional params
                                ref='my_trade_ref',
                                stop_relative=10,
                                limit_relative=20,
                                meta={'data': 'my_data'},
                            )
            ```
        """
        self.market.trade_manager.open_trade(strategy=self, **kwargs)

    def close_trade_by_ref(self, ref: str, **kwargs) -> None:
        """
        Use this method in your closing strategy methods to close a trade by its
        reference.

        !!! info
            Closing strategy method where it is advised to call this method are:

            - `Strategy.on_new_tick_closing_strategy`
            - `Strategy.on_new_candle_closing_strategy`

        !!! seealso
            This method calls `estrade.TradeManager.close_trade_by_ref` method
            and inject this strategy as an argument.

        Arguments:
            ref: existing `estrade.trade.Trade` ref code attached to this strategy on
                current Market.trade_manager
            kwargs: args to close a trade
                (see `estrade.TradeManager.close_trade_by_ref`)
        hello
        """
        self.market.trade_manager.close_trade_by_ref(ref=ref, strategy=self, **kwargs)

    def close_all_trades(self, *args, **kwargs) -> None:
        """
        Use this method in your closing strategy methods to close all opened trades
        for strategy

        !!! info
            Closing strategy method where it is advised to call this method are:

            - `Strategy.on_new_tick_closing_strategy`
            - `Strategy.on_new_candle_closing_strategy`

        !!! seealso
            This method calls `estrade.TradeManager.close_all_trades` method and inject
            this strategy as an argument.

        Arguments:
            args: args to close a trade (see `estrade.TradeManager.close_all_trades`)
            kwargs: args to close a trade (see `estrade.TradeManager.close_all_trades`)
        """
        kwargs.pop('strategy', None)
        self.market.trade_manager.close_all_trades(*args, strategy=self, **kwargs)

    def close_all_buys(self, *args, **kwargs) -> None:
        """
        Use this method in closing strategy methods to close all opened buys for
        strategy

        !!! info
            Closing strategy method where it is advised to call this method are:

            - `Strategy.on_new_tick_closing_strategy`
            - `Strategy.on_new_candle_closing_strategy`

        !!! seealso
            This method calls `estrade.TradeManager.close_all_buys` method and inject
            this strategy as an argument.

        Arguments:
            args: args to close a trade (see `estrade.TradeManager.close_all_buys`)
            kwargs: args to close a trade (see `estrade.TradeManager.close_all_buys`)
        """
        kwargs.pop('strategy', None)
        self.market.trade_manager.close_all_buys(*args, strategy=self, **kwargs)

    def close_all_sells(self, *args, **kwargs) -> None:
        """
        Use this method in closing strategy methods to close all opened sells for
        strategy

        !!! info
            Closing strategy method where it is advised to call this method are:

            - `Strategy.on_new_tick_closing_strategy`
            - `Strategy.on_new_candle_closing_strategy`

        !!! seealso
            This method calls `estrade.TradeManager.close_all_sells` method and inject
            this strategy as an argument.

        Arguments:
            args: args to close a trade (see `estrade.TradeManager.close_all_sells`)
            kwargs: args to close a trade (see `estrade.TradeManager.close_all_sells`)
        """
        kwargs.pop('strategy', None)
        self.market.trade_manager.close_all_sells(*args, strategy=self, **kwargs)

    ##################################################
    # Check if strategy method has to be applied
    ##################################################
    def _apply_closing_strategy(self):
        """
        Check if closing strategy must be applied.
        Closing strategy is applied if trades are opened.
        :return: <bool> True if closing strategy have to be applied, else False
        """
        if self.market.trade_manager.strategy_trades[self.ref]["opened"]:
            return True
        return False

    def _apply_opening_strategy(self):
        """
        Check if opening strategy must be applied.
        Opening strategy have to be applied if the number of open trades is inferior to
        the max concurrent trades
        :return: <bool> True if opening strategy have to be applied, else False
        """
        if (
            len(self.market.trade_manager.strategy_trades[self.ref]["opened"])
            < self.max_concurrent_trades
        ):
            return True
        return False

    ##################################################
    # EVENTS
    ##################################################
    def _is_strategy_paused(self, current_datetime: arrow.Arrow) -> bool:
        """
        Check if strategy is paused for a given tick datetime.
        """
        if self.paused_until is None:
            self.set_pause()

        if self.paused_until is None:
            return False
        elif self.paused_until <= current_datetime:
            logger.info('Un-pause strategy %s' % self.ref)
            self.paused_until = None
            return False
        return True

    def on_new_tick(self, tick: Tick) -> None:
        """
        Method triggered when a new tick is received by Market.
        This method check if opening and/or closing srategies have to be applied and
        then call the opening/closing
        methods.

        Arguments:
            tick: new tick instance received

        """
        logger.debug('check if any trade has to be open/closed on new tick')

        if self._apply_closing_strategy():
            logger.debug('apply closing strategy')
            self.on_new_tick_closing_strategy(tick)

            for candle_set in tick.epic.candle_sets:
                if candle_set.new_candle_opened:
                    self.on_new_candle_closing_strategy(candle_set=candle_set)

        if self.stopped is False:
            self.set_stop()
        if self.stopped is True:
            logger.warning('Strategy %s is stopped, it must not be called' % self.ref)
            return

        if self._is_strategy_paused(tick.datetime):
            logger.debug('Strategy %s is paused, does not apply tick.')
            return

        if self._apply_opening_strategy():
            logger.debug('apply opening strategy')
            self.on_new_tick_opening_strategy(tick)

            for candle_set in tick.epic.candle_sets:
                if candle_set.new_candle_opened:
                    logger.debug(
                        "call new candle opening strategy bc "
                        "a new candle was created in %s" % candle_set.timeframe
                    )
                    self.on_new_candle_opening_strategy(candle_set=candle_set)

    # def on_new_candle(self, candle):
    #     # TODO: check if need to be removed
    #     raise NotImplementedError('On new candle not implemented on Strategy')

    ##################################################
    # Strategy apply method to be implemented
    ##################################################
    def set_pause(self):
        """
        This method is called before every opening and closing strategy methods
        execution. Its purpose is to update the `self.paused_until` parameter
        with a datetime (as arrow.Arrow) representing the target date until no
        opening/closing strategy method will be called.

        When a tick with a datetime superior to the `self.paused_until` value will
        be received from provider, the strategy will be un-paused automatically (the
        `self.paused_until` value will be reset to `None`)

        !!! note:
            Setting the `self.paused_until` does not close trades automatically,
            if required you have to explicitly close trades.

        """
        pass

    def set_stop(self):
        """
        This method is called before every opening and closing strategy methods
        execution. Its purpose is to update the `self.stopped` parameter
        with `True` when the strategy have to be totally stopped.

        !!! note:
            When the `self.stopped` value is set to `True` the strategy is totally
            deactivated. No tick will every be sent again to this strategy.

        !!! note:
            If all strategy are stopped, the market run will automatically end.

        """
        pass

    def on_new_tick_opening_strategy(self, tick: Tick) -> None:
        """
        This method is called on every tick received from provider when the number of
        open trades for this strategy
        is inferior to this instance `max_concurrent_trades` value.

        Arguments:
            tick: new tick received
        """
        pass

    def on_new_tick_closing_strategy(self, tick: Tick) -> None:
        """
        Method is called on every tick received from provider when at least one trade is
        opened for this strategy.

        Arguments:
            tick: new tick received

        """
        pass

    def on_new_candle_opening_strategy(self, candle_set: CandleSet) -> None:
        """
        Method called every time a new candle is created when the number of opened
        trades for this strategy is inferior to this instance `max_concurrent_trades`
        value.

        Use this method to open trades with `Strategy.open_trade` method.

        Arguments:
            candle_set: candle set that generated a new candle

        """
        pass

    def on_new_candle_closing_strategy(self, candle_set: CandleSet) -> None:
        """
        Method called every time a new candle is created and at least one trade is
        opened for this strategy.

        Arguments:
            candle_set: candle set that generated a new candle

        """
        pass

    ##################################################
    # Reporting
    ##################################################
    @property
    def to_json(self) -> Dict[str, Any]:
        """
        Generate a dictionary of strategy stats for reporting
        """
        strategy_stats = {
            "ref": self.ref,
            "max_concurrent_trades": self.max_concurrent_trades,
            "profit_factor": None,
            "result": None,
            "nb_trades": 0,
            "nb_win": 0,
            "nb_losses": 0,
            **self.meta,
        }
        if self.market and self.market.trade_manager:
            strategy_stats["profit_factor"] = self.market.trade_manager.profit_factor(
                strategy=self
            )
            strategy_stats["result"] = self.market.trade_manager.result(strategy=self)
            strategy_stats["nb_trades"] = self.market.trade_manager.nb_trades(
                strategy=self
            )
            strategy_stats["nb_win"] = self.market.trade_manager.nb_win(strategy=self)
            strategy_stats["nb_losses"] = self.market.trade_manager.nb_losses(
                strategy=self
            )
            strategy_stats["max_drawdown"] = self.market.trade_manager.max_drowdown(
                strategy=self
            )

        return strategy_stats

    @property
    def json_headers(self) -> List[str]:
        """
        Generate header of a dictionary of strategy stats for reporting
        """
        headers = [
            "ref",
            "profit_factor",
            "result",
            "nb_trades",
            "nb_win",
            "nb_losses",
            "max_drawdown",
            "max_concurrent_trades",
        ]
        headers.extend(list(self.meta.keys()))
        return headers



class EstradeException(Exception):
    message_prefix = ''

    def __init__(self, message, *args, **kwargs):
        message = f'Estrade::{self.message_prefix}::{message}'
        super().__init__(message, *args, **kwargs)


class MarketException(EstradeException):
    message_prefix = 'Market'


class EpicException(EstradeException):
    message_prefix = 'Epic'


class CandleSetException(EstradeException):
    message_prefix = 'CandleSet'


class CandleException(EstradeException):
    message_prefix = 'Candle'


class ProviderException(EstradeException):
    message_prefix = 'Provider'


class ReportingException(EstradeException):
    message_prefix = 'Reporting'


class StopLimitException(EstradeException):
    message_prefix = 'StopLimit'


class StrategyException(EstradeException):
    message_prefix = 'Strategy'


class TickException(EstradeException):
    message_prefix = 'Tick'


class TradeManagerException(EstradeException):
    message_prefix = 'TradeManager'


class TradeException(EstradeException):
    message_prefix = 'Trade'



import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from estrade.market import Market

logger = logging.getLogger(__name__)


class ReportingMixin:
    """
    Define Reporting Abstract class.
    A reporting class is able to generate reporting:
        - on new tick
        - on trade update (open/update/close)
        - on epic update
        - when all ticks are generated
    """

    def __init__(self, market: 'Market') -> None:
        self.market = market
        self.market.reporting = self
        self.subscribe()

    def subscribe(self):
        """
        subscribe Reporting class to events depending if methods were implemented.
        :return:
        """
        if self.on_new_tick != ReportingMixin.on_new_tick:
            logger.info(
                'subscribe Reporting %s to on_new_tick events' % self.__class__.__name__
            )
            for epic in self.market.epics:
                self.market.subscribe(
                    f'market_after_on_new_tick_{epic.ref}', self.on_new_tick
                )

        if self.on_trade_update != ReportingMixin.on_trade_update:
            # TODO: implement events in trade_manager
            logger.info(
                'subscribe Reporting %s to on_trade_update events'
                % self.__class__.__name__
            )
            self.market.trade_manager.subscribe('trade_update', self.on_trade_update)

        if self.on_run_end != ReportingMixin.on_run_end:
            logger.info(
                'subscribe Reporting %s to on_new_tick events' % self.__class__.__name__
            )
            self.market.subscribe('market_run_end', self.on_run_end)

    def on_new_tick(self, tick):
        """
        Method called when a new tick is received by market
        :param tick: <estrade.tick.Tick>
        :return:
        """
        raise NotImplementedError(
            'To report on every tick, please implement your Reporting '
        )

    def on_trade_update(self, trade):
        """
        Method called when a trade is created/updated/closed
        # TODO: implement events in trade_manager
        :param trade: <estrade.trade.Trade>
        :return:
        """
        raise NotImplementedError(
            'To report on every epic update, please implement your Reporting '
        )

    def on_run_end(self):
        """
        Method called when Market run ends.
        :return:
        """
        raise NotImplementedError(
            'To report on every epic update, please implement your Reporting '
        )



"""This file define the CandleSet object.
"""
import logging
import re
from datetime import datetime, timedelta
from typing import List, Optional, Tuple, TYPE_CHECKING

from arrow import Arrow

from estrade.epic import Epic
from estrade.mixins.candle_set_indicator_mixin import CandleSetIndicatorMixin
from estrade.candle import Candle
from estrade.exceptions import CandleSetException
from estrade.observer import Observable

if TYPE_CHECKING:
    from estrade.tick import Tick

logger = logging.getLogger(__name__)

ALLOWED_UT = [
    'ticks',
    'seconds',
    'minutes',
    'hours',
    'days',
]


class CandleSet(Observable):
    """
    A CandleSet create candles, add ticks and close a list of `estrade.Candles`
    based on its timeframe.

    Arguments:
        epic: Epic instance to attach to CandleSet
        timeframe: Define either a number of time units either a number of ticks.
        indicators: List of indicators attached to CandleSet
        max_candles_in_memory: Number of candles kept in memory

    !!! info "timeframe"
        - unit must always be plural ('1hours' and not '1hour')
        - inside a CandleSet timeframe is splitted between:
            - `ut` (time unit, eg. ticks, minutes, hours etc.)
            - `nb` (nb of time unit)
        - eg: '3ticks', '56ticks', '30seconds', '3minutes', '4hours'.

    !!! info "max_candles_in_memory"
        - note: the bigger the number of candles, the slower your program will be
        - warning: keep this number high enough for your indicators using candle
        to work properly
        (eg. if you use an indicator for moving average 200, the max number of candles
        in memory should at least be 200)

    !!! example
        Example of usage of candle set

        ```python
        from estrade import Epic, CandleSet, Market

        market = Market()
        # create candle sets with your timeframes and assign them to your epics
        epic1 = Epic(market=market, ref='EPIC1')
        cs_21ticks = CandleSet(epic=epic1, timeframe='21ticks')
        cs_1mn1 = CandleSet(epic=epic1, timeframe='1minutes')

        epic2 = EPIC(market=market, ref='EPIC2')
        cs_1mn2 = CandleSet(epic=epic2, timeframe='1minutes')
        cs_4hours = CandleSet(epic=epic2, timeframe='4hours')

        ```

    """

    def __init__(
        self,
        epic: Epic,
        timeframe: str,
        indicators: List[CandleSetIndicatorMixin] = None,
        max_candles_in_memory: int = 100,
    ) -> None:
        logger.debug('Create new candle Set with timeframe %s' % timeframe)
        # CandleSet fire events to
        #   - indicators (to update indicator values)
        #   - strategies (on open/close candle)
        Observable.__init__(self)

        # epic is null on init
        # epic is set when CandleSet is set in candlesets of an
        # `estrade.epic.Epic` instance
        if not isinstance(epic, Epic):
            raise CandleSetException('Invalid epic')
        self.epic: Epic = epic
        self.epic.candle_sets.append(self)

        # ut and nb are set by timeframe setter
        self.ut: Optional[str] = None
        self.nb: Optional[int] = None
        self.timeframe = timeframe

        # init an empty list of candles
        self.candles: List[Candle] = []

        # add candleSet indicators
        self.indicators = indicators

        self.max_candles_in_memory = max_candles_in_memory
        logger.debug('Candle set created with timeframe %s' % self.timeframe)

    ##################################################
    # REF
    ##################################################
    @property
    def ref(self) -> str:
        """
        Returns:
            the reference (~name) of a Candle Set.

        !!! info
            Ref of a candle set cannot be set manually, it is a fixed name with
            format `<epic_ref>-<nb><ut>`

            example: NASDAQ-4hours

        !!! note
            CandleSet has a ref `undefined` while no epic is set.

        """
        return f'{self.epic.ref}-{self.nb}{self.ut}'

    ##################################################
    # TIMEFRAME -> UT + NB
    ##################################################
    @property
    def timeframe(self) -> str:
        """
        Returns:
            the timeframe value as provided in CandleSet instanciation.

        """
        return self._timeframe

    @timeframe.setter
    def timeframe(self, timeframe: str) -> None:
        """
        Split timeframe value provided in instance init to define:
            - ut: time unit name, that can be any value of ALLOWED_UT
            - nb: number of time units, that is a int
        :param timeframe: <str>
        :return:
        """
        nb, ut = self.split_timeframe(timeframe)

        if self.check_timeframe_coherence(ut=ut, nb=nb):
            self.ut = ut
            self.nb = nb
            self._timeframe: str = timeframe

    @staticmethod
    def validate_timeframe(timeframe: str) -> bool:
        nb, ut = CandleSet.split_timeframe(timeframe)
        if CandleSet.check_timeframe_coherence(ut, nb):
            return True
        return False

    @staticmethod
    def split_timeframe(timeframe: str) -> Tuple[int, str]:
        regex_result = re.search(r'^(\d+)(\w+)$', timeframe)
        if not regex_result:
            raise CandleSetException(f'Impossible to create candle set {timeframe}')
        ut = regex_result.group(2).lower()
        nb = int(regex_result.group(1))

        if ut.lower() not in ALLOWED_UT:
            raise CandleSetException('Invalid timeframe UT')
        try:
            nb = int(nb)
        except ValueError:
            raise CandleSetException('Invalid timeframe nb')

        if nb <= 0:
            raise CandleSetException('Invalid timeframe NB value')

        return nb, ut

    @staticmethod
    def check_timeframe_coherence(ut: str, nb: int) -> bool:
        """
        Check that ut and nb (result of timeframe split) are coherent.

         - if ut is in minutes/seconds, nb must be inferior to 60 and nb/60 must be
         an integer.
         - if ut is in hours, nb must be inferior to 12 and nb/12 must be an integer.
         - as only one day period is supported: if ut is 'days' nb must be 1.

        !!! note
            nb and ut values individual consistencies are checked in ut and nb setters.

        """
        if ut in ['seconds', 'minutes'] and int(60 / nb) != 60 / nb:
            raise CandleSetException(
                f'Inconsistent UT: munutes/seconds timeframe must be '
                f'inferior to 60 and {nb}/60 must be an integer.'
            )
        elif ut in ['hours'] and int(12 / nb) != 12 / nb:
            raise CandleSetException(
                f'Inconsistent UT: hour timeframe must be inferor or equal to 12. '
                f'And {nb}/12 must be an int'
            )
        elif ut == 'days' and nb != 1:
            raise CandleSetException(
                'Inconsistent UT, only 1days timeframe is allowed.'
            )

        return True

    ##################################################
    # INDICATORS
    ##################################################
    @property
    def indicators(self) -> Optional[List[CandleSetIndicatorMixin]]:
        """
        Returns:
            Return list of indicators attached to candle set
        """
        return self._indicators

    @indicators.setter
    def indicators(self, indicators: Optional[List[CandleSetIndicatorMixin]]) -> None:
        """
        Set candle set indicators.
        """
        self._indicators = []

        if indicators:
            if not isinstance(indicators, list):
                raise CandleSetException('Indicators must be a list')

            for indicator in indicators:
                if not isinstance(indicator, CandleSetIndicatorMixin):
                    raise CandleSetException(
                        'Indicators must be instances of '
                        'estrade.Acandle_set_indicator.AbstractCandleSetIndicator'
                    )
                logger.debug('add indicator %s' % indicator.name)
                indicator.candle_set = self
                self._indicators.append(indicator)

    def indicator(self, indicator_name: str) -> Optional[CandleSetIndicatorMixin]:
        """
        Arguments:
            indicator_name: name of the indicator to fetch

        Raises:
            CandleSetException: if no indicator is found with this name

        Returns:
            the an indicator in this instance indicators by its name.
        """
        if self.indicators:
            for indicator in self.indicators:
                if indicator.name == indicator_name:
                    return indicator
        raise CandleSetException(f'No indicator found with name {indicator_name}')

    ##################################################
    # CANDLE GETTERS PROPERTIES
    ##################################################
    @property
    def current_candle(self) -> Optional[Candle]:
        """
        Returns:
            return the candle that is currently fed by ticks received.

        """
        return self.candles[-1] if self.candles else None

    @property
    def last_closed_candle(self) -> Optional[Candle]:
        """
        Returns:
            return the last candle that was closed.

        !!! info
            At beginning of Market run, the ticks feed a the first candle,
            while this first candle is not closed (=its timeframe over) the
            `last_closed_candle` is empty.
            As a consequence, in your strategies, it is strongly advised
            that your check that the `last_closed_candle` is not None before using it.
        """
        for c in reversed(self.candles):
            if c.closed:
                return c
        return None

    @property
    def new_candle_opened(self) -> bool:
        """
        Check if current candle is newly opened (only one tick is in the last candle)

        Returns:
            True if the current candle has only one tick.
        """
        return len(self.current_candle.ticks) <= 1 if self.current_candle else False

    @property
    def is_closing_candle(self) -> bool:
        """
        This method check if the last candle will be closed on next tick.

        Returns:
            True if the current candle will be closed on next tick, else False

        Raises:
            CandleSetException: if attempt to use this method on a timed CandleSet.

        !!! warning
            This method can only be used when this instance UT is ticks.
            It cannot be used on time based candlesets.

        """
        if not self.ut == 'ticks':
            raise CandleSetException(
                'Impossible to determine if it is the last tick on timed candles'
            )
        return bool(len(self.candles[-1].ticks) == self.nb)

    @property
    def nb_closed_candles(self) -> int:
        """
        Returns:
            The number of closed candles

        !!! warning
            when the number of closed candle reach the `max_candles_in_memory` value
            set on init, then this value will always returns the `max_candles_in_memory`

        """
        return len(self.candles) - 1 if self.candles else 0

    def closed_candles(self, nb: int = None) -> List[Candle]:
        """
        Get a list of the last closed candles

        Arguments:
            nb: nb of closed candles to retrieve

        """
        if nb and nb > self.nb_closed_candles:
            raise CandleSetException(
                f'Impossible to get {nb} candles from {self.ref} '
                f'bc only contains {self.nb_closed_candles}'
            )
        elif nb is None:
            nb = self.nb_closed_candles

        return self.candles[((nb + 1) * -1) : -1]

    ##################################################
    # CANDLE OPEN/CLOSE METHODS
    ##################################################
    @staticmethod
    def round_candle_open_datetime(tick_datetime: Arrow, nb: int, ut: str) -> Arrow:
        """
        For "timed" candle set (candleSet with an UT that is a time unit.).
        The open time must be rounded down.

        Eg. with a CandleSet having a '5minutes' timeframe. If the first tick received
        to open the candle is 22 min and 03 seconds, it must be rounded to the closest
        down division of 5minutes. In this case, the candle must be opened
        at 20mn and 00seconds.

        Arguments:
            tick_datetime: the tick datetime to round to the timeframe open
            nb: the timeframe number of UT
            ut: the time unit

        Returns:
            Datetime rounded to the timeframe open
        """
        if ut == 'seconds':
            return tick_datetime - timedelta(
                seconds=tick_datetime.second % nb,
                microseconds=tick_datetime.microsecond,
            )
        elif ut == 'minutes':
            return tick_datetime - timedelta(
                minutes=tick_datetime.minute % nb,
                seconds=tick_datetime.second,
                microseconds=tick_datetime.microsecond,
            )
        elif ut == 'hours':
            return tick_datetime - timedelta(
                hours=tick_datetime.hour % nb,
                minutes=tick_datetime.minute,
                seconds=tick_datetime.second,
                microseconds=tick_datetime.microsecond,
            )
        elif ut == 'days':
            return tick_datetime - timedelta(
                days=tick_datetime.day % nb,
                hours=tick_datetime.hour,
                minutes=tick_datetime.minute,
                seconds=tick_datetime.second,
                microseconds=tick_datetime.microsecond,
            )
        return tick_datetime

    def _create_new_candle(self, tick: 'Tick') -> None:
        """
        This method create a new candle from an tick and add it to
        the instance list of candles.
        This method fire candle opening events.

        :param tick: instance of <estrade.tick.Tick>
        :return:
        """
        if self.epic is None:
            raise CandleSetException('Cannot create candle when epic is not set.')

        self.candles.append(
            Candle(timeframe=self.timeframe, epic_ref=self.epic.ref, open_tick=tick)
        )
        self.candles = self.candles[(self.max_candles_in_memory * -1) :]

    def _is_current_candle_finished(self, dt: datetime) -> bool:
        """
        Check if the current candle is finished.
        For CandleSet with a time unit in ticks, it means check if
        the current candle nb of ticks is reached.
        For CandleSet with a "timed" time unit, it means check if
        the received datetime is out of the current candle
        time range.

        note: the input datetime is the tick received datetime.

        :param dt: <datetime.datetime>
        :return: <bool>
        """
        if self.ut is None or self.nb is None:
            raise CandleSetException(
                'Cannot define if candle is finished when' 'ut or nb is not defined.'
            )
        if not self.current_candle:
            raise CandleSetException(
                'Cannot define if candle is finished when no '
                'tick were added to candleSet.'
            )
        if self.ut == 'ticks':
            if len(self.current_candle.ticks) >= self.nb:
                logger.debug('%s: current candle is finished' % self.timeframe)
                return True
        else:
            timedelta_params = {self.ut: self.nb}
            if self.current_candle.open_at + timedelta(**timedelta_params) <= dt:
                logger.debug(
                    '%s: current candle is finished: %s + %s <= %s'
                    % (
                        self.timeframe,
                        self.current_candle.open_at,
                        timedelta_params,
                        dt,
                    )
                )
                return True
            else:
                logger.debug(
                    '%s: current candle is not finished %s + %s < %s'
                    % (
                        self.timeframe,
                        self.current_candle.open_at,
                        timedelta_params,
                        dt,
                    )
                )
        return False

    def _close_last_candle(self) -> None:
        """
        This method close the current candle.
        """
        if self.current_candle:
            self.current_candle.close_candle()

    ##################################################
    # EVENTS
    ##################################################
    def on_new_tick(self, tick: 'Tick') -> None:
        """
        This method is the entry point called on every new tick.

         - update candles (close and open candles if required or add tick
         to current candle)
         - send events (to update indicators, call strategies candle events etc.)

        Arguments:
            tick: new tick received to be dispatched to candles and indicators

        """
        logger.debug('%s : add tick in candle set: %s' % (self.timeframe, tick))
        if not self.epic:
            raise CandleSetException(
                'Impossible to add tick to a candle set with no epic'
            )

        self.fire(f'candle_set_before_on_new_tick_{tick.epic.ref}', tick=tick)
        new_candle_created = False
        current_candle_closed = False
        if not self.current_candle:
            logger.debug('%s : init list of candles with a new candle' % self.timeframe)
            self._create_new_candle(tick)
            new_candle_created = True
        else:
            logger.debug(
                '%s: check if tick datetime %s require to close candle'
                % (self.timeframe, tick.datetime)
            )
            if self._is_current_candle_finished(tick.datetime):
                logger.debug('%s: close candle' % self.timeframe)
                self._close_last_candle()
                current_candle_closed = True

                logger.debug('%s: create a new candle' % self.timeframe)
                self._create_new_candle(tick)
                new_candle_created = True

            else:
                logger.debug('%s: add new tick to current candle' % self.timeframe)
                self.current_candle.on_new_tick(tick)

        # update indicators
        if self.indicators:
            for indicator in self.indicators:

                if current_candle_closed:
                    indicator.on_candle_close(closed_candle=self.candles[-2])
                if new_candle_created:
                    indicator.on_new_candle(new_candle=self.candles[-1])

                indicator.on_new_tick(tick=tick)

        self.fire(f'candle_set_after_on_new_tick_{tick.epic.ref}', tick=tick)



import uuid

from estrade.mixins import get_exception


class RefMixin:
    """
    Abstract class to use as parent of clases having a reference
    """

    def __init__(self, ref):
        self.ref = ref

    @property
    def ref(self):
        return self._ref

    @ref.setter
    def ref(self, ref):
        """
        A reference is either :
            - a string provided on init
            - a string generated as <class_name>_<uuid>[:4]
        :param ref: <str>
        :return:
        """
        if ref is not None:
            if not isinstance(ref, str):
                raise get_exception(self)('Impossible to set a non string as ref')
            self._ref = ref
        else:
            self.ref = f'{self.__class__.__name__.lower()}_{str(uuid.uuid4())[:4]}'



import logging
import pytz
from dateutil import tz
from typing import List, TYPE_CHECKING

from estrade.mixins.ref_mixin import RefMixin
from estrade.exceptions import EpicException

if TYPE_CHECKING:
    from estrade.candle_set import CandleSet
    from estrade.market import Market
    from estrade.strategy import Strategy
    from estrade.tick import Tick

logger = logging.getLogger(__name__)


class Epic(RefMixin):
    """
    Class used to define an Epic

    Arguments:
        ref: reference/name of the epic
        market: Market instance
        timezone: epic timezone
        ticks_in_memory: nb ticks kept in the `ticks` attribute

    !!! example
        ```python
        from estrade import Epic, Market

        market = Market()
        epic = Epic(
            ref='MY_EPIC',
            market=market,
            timezone='Europe/Paris',
            ticks_in_memory=50
        )
        ```

    """

    def __init__(
        self,
        market: 'Market',
        ref: str = None,
        timezone: str = 'UTC',
        ticks_in_memory: int = 1,
    ):
        RefMixin.__init__(self, ref)
        logger.info('Init new Epic %s' % self.ref)

        self.timezone = timezone
        self.strategies: List[Strategy] = []

        # init a market to None
        self.market = market
        self.market.epics.append(self)

        # TODO: check that elements appened to this list are CandleSet objects
        self.candle_sets = []
        self.tradeable = True
        self.meta = {}
        self.ticks_in_memory = ticks_in_memory
        self.ticks = []
        logger.info(
            'New epic created with ref %s and %d strategies'
            % (self.ref, len(self.strategies))
        )

    ##################################################
    # CANDLE SETS
    ##################################################
    def get_candle_set(self, timeframe: str) -> 'CandleSet':
        """
        Get candle_set by timeframe in this instance candle_set

        Arguments:
            timeframe: timeframe of CandleSet to find in epic

        Returns:
            CandleSet associated with the required timeframe

        """
        for cs in self.candle_sets:
            if cs.timeframe == timeframe:
                return cs
        raise EpicException(
            'No candleSet found with timeframe %s in epic %s' % (timeframe, self.ref)
        )

    ##################################################
    # TICKS
    ##################################################
    @property
    def current_tick(self):
        return self.ticks[-1] if self.ticks else None

    def _add_tick(self, tick):
        """
        Add a new tick to list of epic ticks
        :param tick: <estrade.tick.Tick> instance
        :return:
        """
        self.ticks.append(tick)
        self.ticks = self.ticks[(self.ticks_in_memory * -1) :]

    ##################################################
    # TIMEZONE
    ##################################################
    @property
    def timezone(self) -> str:
        """
        Returns:
            epic timezone (see `pytz.all_timezones`)
        """
        return self._timezone

    @timezone.setter
    def timezone(self, timezone: str) -> None:
        """
        Set Epic timezone

        Timezone is used to update tick date when a new `estrade.tick.Tick`
        instance is created.

        :param timezone: <str> that should be a valid pytz timezone.
        :return: EpicException if timezone is invalid
        """
        if timezone in pytz.all_timezones:
            self._timezone = timezone
        else:
            raise EpicException(f'Invalid timezone : {timezone}')

    ##################################################
    # EVENTS
    ##################################################
    def _check_timezone(self, tick: 'Tick') -> bool:
        """Check that epic timezone matches the epic timezone."""
        if (
            self.timezone != 'UTC' and tick.datetime.tzinfo != tz.gettz(self.timezone)
        ) or (self.timezone == 'UTC' and tick.datetime.tzinfo != tz.tzutc()):
            raise EpicException(
                f'Invalid tick timezone, found {tick.datetime.tzinfo} when'
                f' expected {tz.gettz(self.timezone)}'
            )
        return True

    def on_new_tick(self, tick: 'Tick') -> None:
        """
        Method used to handle every new tick received by market.

        Its purpose is to:

         - check the tick received
         - update the epic/max min
         - dispatch tick to CandleSets

        Arguments:
            tick: tick received by epic

        """
        if tick.epic != self:
            raise EpicException(
                f'Invalid tick epic: cannot update epic {self.ref}({id(self)}) with '
                f'tick attached to epic {tick.epic.ref}({id(tick.epic)})'
            )

        logger.debug(
            'Add new tick to epic %s (%s): %f' % (self.ref, id(self), tick.value)
        )
        self._check_timezone(tick)

        # set tick as the current tick
        self._add_tick(tick)

        # dispatch tick to all candle sets
        for candle_set in self.candle_sets:
            logger.debug('dispatch tick to candle set %s' % candle_set.timeframe)
            candle_set.on_new_tick(tick)



import logging

logger = logging.getLogger(__name__)


class Observable:
    """
    Class used to implement observer pattern.
    """

    def __init__(self):
        """
        When instanciated, an observer class create an empty dict of callbacks.
        """
        logger.debug('init %s as observable' % self.__class__.__name__)
        self.callbacks = {}

    def subscribe(self, event_name, callback):
        """
        When another object subscribe to this object, it define the name of event
        and the callback function
        :param event_name: <str>
        :param callback: function name
        :return:
        """
        logger.debug(
            'Add callback %s on to event %s fired by %s'
            % (callback, event_name, self.__class__.__name__)
        )
        if event_name not in self.callbacks:
            self.callbacks[event_name] = [callback]
        elif callback not in self.callbacks[event_name]:
            self.callbacks[event_name].append(callback)
        else:
            logger.warning(
                'callback %s is attached twice to event %s' % (callback, event_name)
            )

    def fire(self, event_name, *args, **kwargs):
        """
        When an event occurs, this function is called with the event name and call
        all defined callbacks with args/kwargs.
        :param event_name: <str>
        :return:
        """
        logger.debug('fire event %s from %s' % (event_name, self.__class__.__name__))
        if not self.callbacks.get(event_name):
            logger.debug(
                'Event %s is not subscribed by any object, '
                'it will not call any callback' % event_name
            )
        else:
            for callback in self.callbacks.get(event_name, []):
                logger.debug('Call callback %s' % callback)
                callback(*args, **kwargs)



import logging
from typing import Optional

logger = logging.getLogger(__name__)


class CandleSetIndicatorMixin:
    def __init__(self, name: str) -> None:
        self.candle_set: Optional['CandleSet'] = None
        self.name = name

    def on_new_tick(self, tick):
        raise NotImplementedError()

    def on_new_candle(self, new_candle):
        raise NotImplementedError()

    def on_candle_close(self, closed_candle):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['estrade',
 'estrade.indicators',
 'estrade.indicators.candle_set',
 'estrade.mixins',
 'estrade.reporting',
 'estrade.utils']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=0.15.5,<0.16.0',
 'pyYAML>=5.3,<6.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pytz>=2019.3,<2020.0',
 'requests>=2.22.0,<3.0.0']

setup_kwargs = {
    'name': 'estrade',
    'version': '0.1.0',
    'description': 'Build, Backtest and Go Live your own trading bots',
    'long_description': '# Estrade\n\n[![Build Status](https://travis-ci.com/cimourdain/estrade.svg?branch=master)](https://travis-ci.com/cimourdain/estrade)\n[![Documentation Status](https://readthedocs.org/projects/estrade/badge/?version=latest)](https://estrade.readthedocs.io/en/latest/?badge=latest)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n\n# Estrade: Trading bot manager\n\nEstrade is a python library that allows you to easily backtest and run stock trading strategies.\n\nEstrade focus on providing tools so you mainly focus on your strategy definition.\n\n>  **WARNING**: Estrade is still in an alpha state of developpement and very unmature. Do not use it for other purposes than testing.\n\n## Features\n\n- Estrade provides a **market environnement**, so you do not have to worry about\n   - Trades result calculation\n   - Candle Graph building\n   - Indicators calculation\n- Estrade allows you to define your strategies based on market events (new tick received, new candle created)\n- Estrade allows you to create your own data providers to generate ticks data and manage trades (open/close)\n- Estrade allows you to create your own indicators\n- Estrade allows you to create your own reporting\n\n\n## What Estrade does NOT provides\n\n- **Data**: You have to define your own data provider (live or static)\n- **Strategies**: Although some very basic (and useless) strategies are provided as examples in samples, Estrate does not provide any financially relevant strategy.\n\n## Documentation\n\n[Documentation](https://estrade.readthedocs.io/)\n\n\n',
    'author': 'Gabriel Oger',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/estrade/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import arrow
from datetime import datetime
from typing import Any, Dict, Optional, Union

from estrade.exceptions import TickException
from estrade.epic import Epic


class Tick:
    """
    Class designed to manage a tick (market point value)

    Arguments:
        epic: epic instance
        datetime: tick datetime
        bid: tick bid value
        ask: tick ask value
        meta: free dictionary where you can store anything you need

    """

    def __init__(
        self,
        epic: 'Epic',
        datetime: Union[datetime, arrow.Arrow],
        bid: Union[float, int],
        ask: Union[float, int],
        meta: Optional[Dict[str, Any]] = None,
    ):

        # check epic
        if not isinstance(epic, Epic):
            raise TickException('Invalid tick epic')
        self.epic = epic

        # check datetime
        if not hasattr(datetime, 'tzinfo') or datetime.tzinfo is None:
            raise TickException('Invalid tick datetime')
        self._datetime = arrow.get(datetime).to(self.epic.timezone)

        # check bid/ask
        if not isinstance(ask, (int, float)):
            raise TickException('Invalid ask value')
        if not isinstance(bid, (int, float)):
            raise TickException('Invalid bid value')
        if bid > ask:
            raise TickException('Inconsistent bid/ask values.')
        self.ask = float(ask)
        self.bid = float(bid)

        self.meta = meta

    @property
    def spread(self):
        """
        The spread is the difference between bid and ask

        :return: (float) spread value
        """
        return round(self.ask - self.bid, 2)

    @property
    def value(self):
        """
        The tick value is the value exactly in the middle between bid and ask

        :return: (float) spread value
        """
        return round(self.bid + (self.spread / 2), 2)

    @property
    def datetime(self) -> arrow.Arrow:
        """
        Returns:
            Tick datetime
        """
        return self._datetime

    @property
    def datetime_utc(self) -> arrow.Arrow:
        """
        Return datetime as UTC, this method is useful when your tick datetime
        is zoned.

        :return: (`arrow.Arrow`): tick datetime zoned to UTC
        """
        return self._datetime.to('UTC')

    def __str__(self) -> str:
        """
        String representation of tick (mainly used for logging)
        :return:
        """
        return (
            f'@{self._datetime}: {self.value} '
            f'(bid: {self.bid}, ask: {self.ask}, spread: {self.spread})'
        )

    def to_json(self, datetime_to_str: bool = False) -> Dict[str, Any]:
        """
        Convert tick to dictionary (mainly used for reporting)
        :param datetime_to_str: <bool> => does the datetime value in dict must convert
        datetime to string?
        :return:
        """
        return {
            'datetime': self._datetime.strftime('%Y-%m-%d %H:%M:%S')
            if datetime_to_str
            else self._datetime,
            'bid': self.bid,
            'ask': self.ask,
            'spread': self.spread,
            'value': self.value,
        }

    @staticmethod
    def validate(tick: 'Tick') -> bool:
        if not isinstance(tick, Tick):
            raise TickException('Invalid tick instance')
        return True



import logging
from datetime import datetime, timedelta
from typing import List, Generator, TYPE_CHECKING

from estrade.candle import Candle
from estrade.exceptions import ProviderException
from estrade.tick import Tick
from estrade.trade import Trade

if TYPE_CHECKING:
    from estrade.market import Market

logger = logging.getLogger(__name__)


class Provider:
    """
    A provider defines how to fetch your source data (ticks or candles).

    The provider `generate` method feed data to the `estrade.Market` object
    (that will dispatch it to candle sets, indicators, strategies etc.)

    Arguments:
        market: market instance
        start_date: datetime.datetime start_date: optional start date
        end_date: datetime.datetime end_date: optional end date
        requires_login: set to True if your provider requires login before market run.

    !!! note
        if you use the `requires_login`, you have to implement a login method.

    !!! example
        ```python
        from estrade import Epic, Market, Provider, Tick

        MyProvider:
            def generate_ticks():
                from tick in my_source:
                    tick = Tick(
                        epic=self.market.get_epic(tick['epic_code']),
                        datetime=tick.datetime,  # must be timezoned
                        bid=tick['bid']
                        ask=tick['ask']
                    )

                    # dispatch tick to market
                    yield tick

        if __name__ == '__main__':
            market = Market()

            epic = Epic(ref='MY_EPIC', market=market)
            # add candlsets if necessary
            # add strategies
            provider = MyProvider(market=market)

            market.run()
        ```
    """

    def __init__(
        self,
        market: 'Market',
        start_date: datetime = None,
        end_date: datetime = None,
        requires_login: bool = False,
    ):
        # init a market property to None
        self.market = market
        self.market.provider = self
        self.market.subscribe('market_before_run', self.login)

        self.start_date = start_date
        self.end_date = end_date

        self.requires_login = requires_login

    # @property
    # def start_date(self):
    #     return self._start_date

    # @start_date.setter
    # def start_date(self, start_date):
    #     self._start_date = None
    #     if isinstance(start_date, str):
    #         try:
    #             self._start_date = datetime.strptime(start_date, '%Y-%m-%d')
    #         except ValueError:
    #             raise ProviderException('Invalid start date format')

    #     if isinstance(start_date, datetime):
    #         self._start_date = start_date

    # @property
    # def end_date(self):
    #     return self._end_date

    # @end_date.setter
    # def end_date(self, end_date):
    #     self._end_date = None
    #     if isinstance(end_date, str):
    #         try:
    #             self._end_date = datetime.strptime(end_date, '%Y-%m-%d')
    #         except ValueError:
    #             raise ProviderException('Invalid start date format')

    #     if isinstance(end_date, datetime):
    #         self._end_date = end_date

    @property
    def dates_generator(self):
        if not self.start_date or not self.end_date:
            raise ProviderException(
                'Cannot use backtest_dates_generator when '
                'either start_date or end_date is not defined'
            )
        for n in range(int((self.end_date - self.start_date).days + 1)):
            trading_date = self.start_date + timedelta(n)
            yield trading_date.date()

    ##################################################
    # EVENTS
    ##################################################
    def generate_ticks(self) -> Generator[Tick, None, None]:
        """
        This method must be implemented to fetch ticks or candles from provider.

        !!! example "Example of a provider generating ticks"

            ```python
            from estrade import Provider
            from estrade import Tick

            class MyProvider(Provider):

                def generate(self):
                    '''
                        This method parse all ticks from your database.
                        It assumes that your ticks holds the following attributes:
                        - epic code
                        - bid
                        - ask
                        - datetime
                    '''
                    # eg. query ticks in your database (ordered by datetime)
                    for tick in my_database:
                        # build a tick from your data
                        tick = Tick(
                            epic=self.market.get_epic(tick['epic_code']),
                            bid=tick['bid'],
                            ask=tick['ask'],
                            datetime['datetime'],
                        )
                    # dispatch tick to market
                    yield tick
            ```
        """
        pass

    def generate_candles(self) -> Generator[Candle, None, None]:
        """
        Implement this method to run in "Candle mode" meaning that the generate method
        feed `estrade.Candle` objects to the market.

        !!! warning
            Using candle mode is not recommended, it should be reserved to exceptional
            cases where you only have candle data.

            In addition, your your candle data must be timed (open, close, high and
                                                              low must be timed).

        !!! example "example of a provider in candle mode"
            ```python
            from estrade import Provider
            from estrade import Candle, Tick

            class MyProvider(Provider):

                def generate(self):
                # eg. query candle in your database (ordered by datetime)
                for candle in my_database:
                    epic = self.market.get_epic(candle['epic_code'])
                    # build an open tick from your data
                    open = Tick(
                        epic=epic,
                        bid=candle['open_bid'],
                        ask=candle['open_ask'],
                        datetime=candle['open_datetime'],
                    )
                    high = Tick(
                        epic=epic,
                        bid=candle['high_bid'],
                        ask=candle['high_ask'],
                        datetime=candle['high_datetime'],
                    )
                    low = Tick(
                        epic=epic,
                        bid=candle['low_bid'],
                        ask=candle['low_ask'],
                        datetime=candle['low_datetime'],
                    )
                    close = Tick(
                        epic=epic,
                        bid=candle['close_bid'],
                        ask=candle['close_ask'],
                        datetime=candle['close_datetime'],
                    )
                    # build a candle with your ticks
                    candle = Candle(
                        timeframe=candle['timeframe'],
                        epic_ref=candle['epic_code'],
                        open_tick=open
                    )
                    # add high, low tick to candle
                    candle.on_new_tick(high)
                    candle.on_new_tick(low)
                    # finally add the close tick to candle
                    candle.on_new_tick(close)

                # dispatch candle to market
                yield(candle)
            ```
        """
        pass

    def get_open_trades(self) -> List[Trade]:
        """
        Implement this method to fetch list of open trades from your provider.

        This method is called before market run to load existing opened trades.

        Returns:
            List of existing trades to load before Market run
        """
        return []

    def login(self) -> bool:
        """
        This method must perform login process with your provider.

        Returns:
            - True if login to provider was successful successfully
            - False if login to provider failed

        """
        return True

    def update_epic_status(self, epic_ref: str, tradeable_status: bool) -> None:
        """
        This method allow to set epic as (not) tradeable.
        When an epic is not tradeable, no open/close trade will be attempted.

        Arguments:
            epic_ref: epic code
            tradeable_status: Is trading allowed on this epic
        """
        epic = self.market.get_epic(epic_ref)
        epic.tradeable = tradeable_status

    def open_trade(self, trade: Trade) -> Trade:
        """
        For a remote provider, use this method to ask your provider to create
        a new trade.

        Arguments:
            trade: trade to be opened

        !!! tip

            - You can use the `trade.meta` property to store any necessary data for
            your provider.
            - In case of error it is recommended to raise a
            `estrade.exception.ProviderException`

        """
        return trade

    def close_trade(self, trade: Trade, quantity: int) -> Trade:
        """
        For a remote provider, use this method to ask your provider to close a trade.

        Arguments:
            trade: trade to be closed
            quantity: quantity to close

        !!! tip

             - You can use the `trade.meta` property to store any necessary data
             for your provider.
             - In case of error it is recommended to raise a
             `estrade.exception.ProviderException`
        """
        return trade

    def update_trade(self, *args, **kwargs) -> None:
        """
        For a remote provider, use this method to update a trade in case your provider
        notifies an update.

        !!! example

            ```python
            from estrade import Provider


            class MyProvider(Provider):
                # ...
                def update_trade(self, remote_update):
                    # search for existing trade by its ref if your provider
                    # update contains it.
                    trade = self.market.trade_manager.get_trade_by_ref(
                        ref=remote_update['ref']
                    )
                    trade.set_stop(value=remote_update['stop'], relative=False)

                    # search in all existing trades if you stored reconciliation data
                    # in trades meta
                    for trade in self.market.trade_manager.trades:
                        if trade.meta['remote_id'] == remote_update['id']:
                            # update your trade
                            trade.set_stop(value=remote_update['stop'], relative=False)
            ```
        """
        pass



from estrade.mixins.stop_limit_mixin import StopLimitMixin
from estrade.exceptions import StopLimitException


class StopLimitMixinAbsolute(StopLimitMixin):
    """
    Class used to manage absolute stop and limits for <estrade.trade.Trade> classes
    """

    ##################################################
    # DIRECTION
    ##################################################
    @property
    def direction(self):
        """
        Not used for absolute stop/limit
        :return: None
        """
        return None

    @property
    def is_absolute(self):
        """
        Check if stop/limit is absolute
        :return: <bool>
        """
        return True

    ##################################################
    # VALUES
    ##################################################
    @property
    def absolute_value(self):
        """
        In an absolute stop/limit the absolute value is the value provided on init
        :return: <int>
        """
        return self._value

    @property
    def relative_value(self):
        """
        calculate relative stop/limit value from trade open and value
        :return:
        """
        if self._value:
            return abs(self.trade.open - self._value)
        return None

    def _check_value_consistency(self, value):
        """
        Check that the value provided is coherent with the trade open and direction
        :param value: <int/float>
        :return:
        """
        if self.trade.direction > 0 and (
            (self.type_ == 'STOP' and value >= self.trade.open)
            or (self.type_ == 'LIMIT' and value <= self.trade.open)
        ):
            raise StopLimitException(
                f'Impossible to set a {self.type_.upper()} ({value}) '
                f'{"higher" if self.type_ == "STOP" else "lower"} than open tick on a '
                f'{"BUY" if self.trade.direction > 0 else "SELL"} opened '
                f'@ {self.trade.open}'
            )
        elif self.trade.direction < 0 and (
            (self.type_ == 'STOP' and value <= self.trade.open)
            or (self.type_ == 'LIMIT' and value >= self.trade.open)
        ):
            raise StopLimitException(
                f'Impossible to set a {self.type_.upper()} ({value}) '
                f'{"higher" if self.type_ == "LIMIT" else "lower"} than open tick on a '
                f'{"BUY" if self.trade.direction > 0 else "SELL"} opened '
                f'@{self.trade.open}'
            )
        return True


class StopLimitMixinRelative(StopLimitMixin):
    """
    This class handles relative stop/limit for trades.
    """

    ##################################################
    # DIRECTION
    ##################################################
    @property
    def direction(self):
        """
        return stop/limit direction
        direction define if stop/limit is above (direction=1) or under (direction=-1)
        the trade open.
        :return: <int> in [-1, 1] or None
        """
        if (self.type_ == 'STOP' and self.trade.direction >= 1) or (
            self.type_ == 'LIMIT' and self.trade.direction <= -1
        ):
            return -1
        elif (self.type_ == 'LIMIT' and self.trade.direction >= 1) or (
            self.type_ == 'STOP' and self.trade.direction <= -1
        ):
            return 1
        return None

    @property
    def is_relative(self):
        """
        Check if stop/limit is relative
        :return: <bool>
        """
        return True

    ##################################################
    # VALUES
    ##################################################
    @property
    def relative_value(self):
        """
        In a relative stop/limit, the relative value is the value provided on init.
        :return: <int> or <float>
        """
        return self._value

    @property
    def absolute_value(self):
        """
        In a relative stop/limit, calculate the absolute value with trade open
        and direction
        :return: <int> or <float>
        """
        if self._value:
            return self.trade.open + (self.direction * self._value)

        return None

    def _check_value_consistency(self, value):
        """
        Override this method that is not applicable for relative.
        :param value: <int> or <float>
        :return:
        """
        pass
